# Copyright 2011 OpenStack Foundation
# Copyright (c) 2015 Rushil Chugh
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests For miscellaneous util methods used with share."""

from unittest import mock

import ddt

from manila.common import constants
from manila.share import utils as share_utils
from manila import test


@ddt.ddt
class ShareUtilsTestCase(test.TestCase):
    def test_extract_host_without_pool(self):
        host = 'Host@Backend'
        self.assertEqual(
            'Host@Backend', share_utils.extract_host(host))

    def test_extract_host_only_return_host(self):
        host = 'Host@Backend'
        self.assertEqual(
            'Host', share_utils.extract_host(host, 'host'))

    def test_extract_host_only_return_pool(self):
        host = 'Host@Backend'
        self.assertIsNone(
            share_utils.extract_host(host, 'pool'))

    def test_extract_host_only_return_backend(self):
        host = 'Host@Backend'
        self.assertEqual(
            'Host@Backend', share_utils.extract_host(host, 'backend'))

    def test_extract_host_missing_backend_and_pool(self):
        host = 'Host'
        # Default level is 'backend'
        self.assertEqual(
            'Host', share_utils.extract_host(host))

    def test_extract_host_only_return_backend_name(self):
        host = 'Host@Backend#Pool'
        self.assertEqual(
            'Backend', share_utils.extract_host(host, 'backend_name'))

    def test_extract_host_only_return_backend_name_index_error(self):
        host = 'Host#Pool'

        self.assertRaises(IndexError,
                          share_utils.extract_host,
                          host, 'backend_name')

    def test_extract_host_missing_backend(self):
        host = 'Host#Pool'
        self.assertEqual(
            'Host', share_utils.extract_host(host))
        self.assertEqual(
            'Host', share_utils.extract_host(host, 'host'))

    def test_extract_host_missing_backend_only_return_backend(self):
        host = 'Host#Pool'
        self.assertEqual(
            'Host', share_utils.extract_host(host, 'backend'))

    def test_extract_host_missing_backend_only_return_pool(self):
        host = 'Host#Pool'
        self.assertEqual(
            'Pool', share_utils.extract_host(host, 'pool'))
        self.assertEqual(
            'Pool', share_utils.extract_host(host, 'pool', True))

    def test_extract_host_missing_pool(self):
        host = 'Host@Backend'
        self.assertIsNone(
            share_utils.extract_host(host, 'pool'))

    def test_extract_host_missing_pool_use_default_pool(self):
        host = 'Host@Backend'
        self.assertEqual(
            '_pool0', share_utils.extract_host(host, 'pool', True))

    def test_extract_host_with_default_pool(self):
        host = 'Host'
        # Default_pool_name doesn't work for level other than 'pool'
        self.assertEqual(
            'Host', share_utils.extract_host(host, 'host', True))
        self.assertEqual(
            'Host', share_utils.extract_host(host, 'host', False))
        self.assertEqual(
            'Host', share_utils.extract_host(host, 'backend', True))
        self.assertEqual(
            'Host', share_utils.extract_host(host, 'backend', False))

    def test_extract_host_with_pool(self):
        host = 'Host@Backend#Pool'
        self.assertEqual(
            'Host@Backend', share_utils.extract_host(host))
        self.assertEqual(
            'Host', share_utils.extract_host(host, 'host'))
        self.assertEqual(
            'Host@Backend', share_utils.extract_host(host, 'backend'),)
        self.assertEqual(
            'Pool', share_utils.extract_host(host, 'pool'))
        self.assertEqual(
            'Pool', share_utils.extract_host(host, 'pool', True))

    def test_append_host_with_host_and_pool(self):
        host = 'Host'
        pool = 'Pool'
        expected = 'Host#Pool'
        self.assertEqual(expected,
                         share_utils.append_host(host, pool))

    def test_append_host_with_host(self):
        host = 'Host'
        pool = None
        expected = 'Host'
        self.assertEqual(expected,
                         share_utils.append_host(host, pool))

    def test_append_host_with_pool(self):
        host = None
        pool = 'pool'
        expected = None
        self.assertEqual(expected,
                         share_utils.append_host(host, pool))

    def test_append_host_with_no_values(self):
        host = None
        pool = None
        expected = None
        self.assertEqual(expected,
                         share_utils.append_host(host, pool))

    def test_get_active_replica_success(self):
        replica_list = [{'id': '123456',
                         'replica_state': constants.REPLICA_STATE_IN_SYNC},
                        {'id': '654321',
                         'replica_state': constants.REPLICA_STATE_ACTIVE},
                        ]
        replica = share_utils.get_active_replica(replica_list)
        self.assertEqual('654321', replica['id'])

    def test_get_active_replica_not_exist(self):
        replica_list = [{'id': '123456',
                         'replica_state': constants.REPLICA_STATE_IN_SYNC},
                        {'id': '654321',
                         'replica_state': constants.REPLICA_STATE_OUT_OF_SYNC},
                        ]
        replica = share_utils.get_active_replica(replica_list)
        self.assertIsNone(replica)


class NotifyUsageTestCase(test.TestCase):
    @mock.patch('manila.share.utils._usage_from_share')
    @mock.patch('manila.share.utils.CONF')
    @mock.patch('manila.share.utils.rpc')
    def test_notify_about_share_usage(self, mock_rpc, mock_conf, mock_usage):
        mock_conf.host = 'host1'
        output = share_utils.notify_about_share_usage(mock.sentinel.context,
                                                      mock.sentinel.share,
                                                      mock.sentinel.
                                                      share_instance,
                                                      'test_suffix')
        self.assertIsNone(output)
        mock_usage.assert_called_once_with(mock.sentinel.share,
                                           mock.sentinel.share_instance)
        mock_rpc.get_notifier.assert_called_once_with('share',
                                                      'host1')
        mock_rpc.get_notifier.return_value.info.assert_called_once_with(
            mock.sentinel.context,
            'share.test_suffix',
            mock_usage.return_value)

    @mock.patch('manila.share.utils._usage_from_share')
    @mock.patch('manila.share.utils.CONF')
    @mock.patch('manila.share.utils.rpc')
    def test_notify_about_share_usage_with_kwargs(self, mock_rpc, mock_conf,
                                                  mock_usage):
        mock_conf.host = 'host1'
        output = share_utils.notify_about_share_usage(mock.sentinel.context,
                                                      mock.sentinel.share,
                                                      mock.sentinel.
                                                      share_instance,
                                                      'test_suffix',
                                                      extra_usage_info={
                                                          'a': 'b', 'c': 'd'},
                                                      host='host2')
        self.assertIsNone(output)
        mock_usage.assert_called_once_with(mock.sentinel.share,
                                           mock.sentinel.share_instance,
                                           a='b', c='d')
        mock_rpc.get_notifier.assert_called_once_with('share',
                                                      'host2')
        mock_rpc.get_notifier.return_value.info.assert_called_once_with(
            mock.sentinel.context,
            'share.test_suffix',
            mock_usage.return_value)



# Copyright (c) 2014 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" GlusterFS volume mapped share layout testcases.
"""

import re
import shutil
import tempfile
from unittest import mock

import ddt
from oslo_config import cfg

from manila.common import constants
from manila import context
from manila import exception
from manila.share import configuration as config
from manila.share.drivers.glusterfs import common
from manila.share.drivers.glusterfs import layout_volume
from manila import test
from manila.tests import fake_utils


CONF = cfg.CONF


def new_share(**kwargs):
    share = {
        'id': 'fakeid',
        'name': 'fakename',
        'size': 1,
        'share_proto': 'glusterfs',
    }
    share.update(kwargs)
    return share


def glusterXMLOut(**kwargs):

    template = """<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<cliOutput>
  <opRet>%(ret)d</opRet>
  <opErrno>%(errno)d</opErrno>
  <opErrstr>fake error</opErrstr>
</cliOutput>"""

    return template % kwargs, ''


FAKE_UUID1 = '11111111-1111-1111-1111-111111111111'
FAKE_UUID2 = '22222222-2222-2222-2222-222222222222'


@ddt.ddt
class GlusterfsVolumeMappedLayoutTestCase(test.TestCase):
    """Tests GlusterfsVolumeMappedLayout."""

    def setUp(self):
        super(GlusterfsVolumeMappedLayoutTestCase, self).setUp()
        fake_utils.stub_out_utils_execute(self)
        self._execute = fake_utils.fake_execute
        self._context = context.get_admin_context()

        self.glusterfs_target1 = 'root@host1:/gv1'
        self.glusterfs_target2 = 'root@host2:/gv2'
        self.glusterfs_server1 = 'root@host1'
        self.glusterfs_server2 = 'root@host2'
        self.glusterfs_server1_volumes = 'manila-share-1-1G\nshare1'
        self.glusterfs_server2_volumes = 'manila-share-2-2G\nshare2'
        self.share1 = new_share(
            export_location=self.glusterfs_target1,
            status=constants.STATUS_AVAILABLE)
        self.share2 = new_share(
            export_location=self.glusterfs_target2,
            status=constants.STATUS_AVAILABLE)
        gmgr = common.GlusterManager
        self.gmgr1 = gmgr(self.glusterfs_server1, self._execute, None, None,
                          requires={'volume': False})
        self.gmgr2 = gmgr(self.glusterfs_server2, self._execute, None, None,
                          requires={'volume': False})
        self.glusterfs_volumes_dict = (
            {'root@host1:/manila-share-1-1G': {'size': 1},
             'root@host2:/manila-share-2-2G': {'size': 2}})
        self.glusterfs_used_vols = set([
            'root@host1:/manila-share-1-1G',
            'root@host2:/manila-share-2-2G'])

        CONF.set_default('glusterfs_servers',
                         [self.glusterfs_server1, self.glusterfs_server2])
        CONF.set_default('glusterfs_server_password',
                         'fake_password')
        CONF.set_default('glusterfs_path_to_private_key',
                         '/fakepath/to/privatekey')
        CONF.set_default('glusterfs_volume_pattern',
                         r'manila-share-\d+-#{size}G$')
        CONF.set_default('driver_handles_share_servers', False)

        self.fake_driver = mock.Mock()
        self.mock_object(self.fake_driver, '_execute',
                         self._execute)
        self.fake_driver.GLUSTERFS_VERSION_MIN = (3, 6)

        self.fake_conf = config.Configuration(None)
        self.mock_object(tempfile, 'mkdtemp',
                         mock.Mock(return_value='/tmp/tmpKGHKJ'))
        self.mock_object(common.GlusterManager, 'make_gluster_call')

        self.fake_private_storage = mock.Mock()

        with mock.patch.object(layout_volume.GlusterfsVolumeMappedLayout,
                               '_glustermanager',
                               side_effect=[self.gmgr1, self.gmgr2]):
            self._layout = layout_volume.GlusterfsVolumeMappedLayout(
                self.fake_driver, configuration=self.fake_conf,
                private_storage=self.fake_private_storage)
        self._layout.glusterfs_versions = {self.glusterfs_server1: ('3', '6'),
                                           self.glusterfs_server2: ('3', '7')}
        self.addCleanup(fake_utils.fake_execute_set_repliers, [])
        self.addCleanup(fake_utils.fake_execute_clear_log)

    @ddt.data({"test_kwargs": {}, "requires": {"volume": True}},
              {"test_kwargs": {'req_volume': False},
               "requires": {"volume": False}})
    @ddt.unpack
    def test_glustermanager(self, test_kwargs, requires):
        fake_obj = mock.Mock()
        self.mock_object(common, 'GlusterManager',
                         mock.Mock(return_value=fake_obj))

        ret = self._layout._glustermanager(self.glusterfs_target1,
                                           **test_kwargs)

        common.GlusterManager.assert_called_once_with(
            self.glusterfs_target1, self._execute,
            self._layout.configuration.glusterfs_path_to_private_key,
            self._layout.configuration.glusterfs_server_password,
            requires=requires)
        self.assertEqual(fake_obj, ret)

    def test_compile_volume_pattern(self):
        volume_pattern = r'manila-share-\d+-(?P<size>\d+)G$'

        ret = self._layout._compile_volume_pattern()

        self.assertEqual(re.compile(volume_pattern), ret)

    @ddt.data({'root@host1:/manila-share-1-1G': 'NONE',
               'root@host2:/manila-share-2-2G': None},
              {'root@host1:/manila-share-1-1G': FAKE_UUID1,
               'root@host2:/manila-share-2-2G': None},
              {'root@host1:/manila-share-1-1G': 'foobarbaz',
               'root@host2:/manila-share-2-2G': FAKE_UUID2},
              {'root@host1:/manila-share-1-1G': FAKE_UUID1,
               'root@host2:/manila-share-2-2G': FAKE_UUID2})
    def test_fetch_gluster_volumes(self, sharemark):
        vol1_qualified = 'root@host1:/manila-share-1-1G'
        gmgr_vol1 = common.GlusterManager(vol1_qualified)
        gmgr_vol1.get_vol_option = mock.Mock(
            return_value=sharemark[vol1_qualified])
        vol2_qualified = 'root@host2:/manila-share-2-2G'
        gmgr_vol2 = common.GlusterManager(vol2_qualified)
        gmgr_vol2.get_vol_option = mock.Mock(
            return_value=sharemark[vol2_qualified])
        self.mock_object(
            self.gmgr1, 'gluster_call',
            mock.Mock(return_value=(self.glusterfs_server1_volumes, '')))
        self.mock_object(
            self.gmgr2, 'gluster_call',
            mock.Mock(return_value=(self.glusterfs_server2_volumes, '')))
        _glustermanager_calls = (self.gmgr1, gmgr_vol1, self.gmgr2, gmgr_vol2)
        self.mock_object(self._layout, '_glustermanager',
                         mock.Mock(side_effect=_glustermanager_calls))
        expected_output = {}
        for q, d in self.glusterfs_volumes_dict.items():
            if sharemark[q] not in (FAKE_UUID1, FAKE_UUID2):
                expected_output[q] = d

        ret = self._layout._fetch_gluster_volumes()

        test_args = ('volume', 'list')
        self.gmgr1.gluster_call.assert_called_once_with(*test_args,
                                                        log=mock.ANY)
        self.gmgr2.gluster_call.assert_called_once_with(*test_args,
                                                        log=mock.ANY)
        gmgr_vol1.get_vol_option.assert_called_once_with(
            'user.manila-share')
        gmgr_vol2.get_vol_option.assert_called_once_with(
            'user.manila-share')
        self.assertEqual(expected_output, ret)

    def test_fetch_gluster_volumes_no_filter_used(self):
        vol1_qualified = 'root@host1:/manila-share-1-1G'
        gmgr_vol1 = common.GlusterManager(vol1_qualified)
        gmgr_vol1.get_vol_option = mock.Mock()
        vol2_qualified = 'root@host2:/manila-share-2-2G'
        gmgr_vol2 = common.GlusterManager(vol2_qualified)
        gmgr_vol2.get_vol_option = mock.Mock()
        self.mock_object(
            self.gmgr1, 'gluster_call',
            mock.Mock(return_value=(self.glusterfs_server1_volumes, '')))
        self.mock_object(
            self.gmgr2, 'gluster_call',
            mock.Mock(return_value=(self.glusterfs_server2_volumes, '')))
        _glustermanager_calls = (self.gmgr1, gmgr_vol1, self.gmgr2, gmgr_vol2)
        self.mock_object(self._layout, '_glustermanager',
                         mock.Mock(side_effect=_glustermanager_calls))
        expected_output = self.glusterfs_volumes_dict

        ret = self._layout._fetch_gluster_volumes(filter_used=False)

        test_args = ('volume', 'list')
        self.gmgr1.gluster_call.assert_called_once_with(*test_args,
                                                        log=mock.ANY)
        self.gmgr2.gluster_call.assert_called_once_with(*test_args,
                                                        log=mock.ANY)
        self.assertFalse(gmgr_vol1.get_vol_option.called)
        self.assertFalse(gmgr_vol2.get_vol_option.called)
        self.assertEqual(expected_output, ret)

    def test_fetch_gluster_volumes_no_keymatch(self):
        vol1_qualified = 'root@host1:/manila-share-1'
        gmgr_vol1 = common.GlusterManager(vol1_qualified)
        gmgr_vol1.get_vol_option = mock.Mock(return_value=None)
        self._layout.configuration.glusterfs_servers = [self.glusterfs_server1]
        self.mock_object(
            self.gmgr1, 'gluster_call',
            mock.Mock(return_value=('manila-share-1', '')))
        _glustermanager_calls = (self.gmgr1, gmgr_vol1)
        self.mock_object(self._layout, '_glustermanager',
                         mock.Mock(side_effect=_glustermanager_calls))
        self.mock_object(self._layout, 'volume_pattern',
                         re.compile(r'manila-share-\d+(-(?P<size>\d+)G)?$'))
        expected_output = {'root@host1:/manila-share-1': {'size': None}}

        ret = self._layout._fetch_gluster_volumes()

        test_args = ('volume', 'list')
        self.gmgr1.gluster_call.assert_called_once_with(*test_args,
                                                        log=mock.ANY)
        self.assertEqual(expected_output, ret)

    def test_fetch_gluster_volumes_error(self):
        test_args = ('volume', 'list')

        def raise_exception(*args, **kwargs):
            if(args == test_args):
                raise exception.GlusterfsException()

        self._layout.configuration.glusterfs_servers = [self.glusterfs_server1]
        self.mock_object(self.gmgr1, 'gluster_call',
                         mock.Mock(side_effect=raise_exception))
        self.mock_object(self._layout, '_glustermanager',
                         mock.Mock(return_value=self.gmgr1))
        self.mock_object(layout_volume.LOG, 'error')

        self.assertRaises(exception.GlusterfsException,
                          self._layout._fetch_gluster_volumes)

        self.gmgr1.gluster_call.assert_called_once_with(*test_args,
                                                        log=mock.ANY)

    def test_do_setup(self):
        self._layout.configuration.glusterfs_servers = [self.glusterfs_server1]
        self.mock_object(self.gmgr1, 'get_gluster_version',
                         mock.Mock(return_value=('3', '6')))
        self.mock_object(self._layout, '_glustermanager',
                         mock.Mock(return_value=self.gmgr1))
        self.mock_object(self._layout, '_fetch_gluster_volumes',
                         mock.Mock(return_value=self.glusterfs_volumes_dict))
        self.mock_object(self._layout, '_check_mount_glusterfs')
        self._layout.gluster_used_vols = self.glusterfs_used_vols
        self.mock_object(layout_volume.LOG, 'warning')

        self._layout.do_setup(self._context)

        self._layout._fetch_gluster_volumes.assert_called_once_with(
            filter_used=False)
        self._layout._check_mount_glusterfs.assert_called_once_with()
        self.gmgr1.get_gluster_version.assert_called_once_with()

    def test_do_setup_unsupported_glusterfs_version(self):
        self._layout.configuration.glusterfs_servers = [self.glusterfs_server1]
        self.mock_object(self.gmgr1, 'get_gluster_version',
                         mock.Mock(return_value=('3', '5')))
        self.mock_object(self._layout, '_glustermanager',
                         mock.Mock(return_value=self.gmgr1))

        self.assertRaises(exception.GlusterfsException,
                          self._layout.do_setup, self._context)

        self.gmgr1.get_gluster_version.assert_called_once_with()

    @ddt.data(exception.GlusterfsException, RuntimeError)
    def test_do_setup_get_gluster_version_fails(self, exc):
        def raise_exception(*args, **kwargs):
            raise exc

        self._layout.configuration.glusterfs_servers = [self.glusterfs_server1]
        self.mock_object(self.gmgr1, 'get_gluster_version',
                         mock.Mock(side_effect=raise_exception))
        self.mock_object(self._layout, '_glustermanager',
                         mock.Mock(return_value=self.gmgr1))

        self.assertRaises(exc, self._layout.do_setup, self._context)

        self.gmgr1.get_gluster_version.assert_called_once_with()

    def test_do_setup_glusterfs_no_volumes_provided_by_backend(self):
        self._layout.configuration.glusterfs_servers = [self.glusterfs_server1]
        self.mock_object(self.gmgr1, 'get_gluster_version',
                         mock.Mock(return_value=('3', '6')))
        self.mock_object(self._layout, '_glustermanager',
                         mock.Mock(return_value=self.gmgr1))
        self.mock_object(self._layout, '_fetch_gluster_volumes',
                         mock.Mock(return_value={}))

        self.assertRaises(exception.GlusterfsException,
                          self._layout.do_setup, self._context)

        self._layout._fetch_gluster_volumes.assert_called_once_with(
            filter_used=False)

    def test_share_manager(self):
        self.mock_object(self._layout, '_glustermanager',
                         mock.Mock(return_value=self.gmgr1))
        self.mock_object(self._layout.private_storage,
                         'get', mock.Mock(return_value='host1:/gv1'))

        ret = self._layout._share_manager(self.share1)

        self._layout.private_storage.get.assert_called_once_with(
            self.share1['id'], 'volume')
        self._layout._glustermanager.assert_called_once_with('host1:/gv1')
        self.assertEqual(self.gmgr1, ret)

    def test_share_manager_no_privdata(self):
        self.mock_object(self._layout.private_storage,
                         'get', mock.Mock(return_value=None))

        ret = self._layout._share_manager(self.share1)

        self._layout.private_storage.get.assert_called_once_with(
            self.share1['id'], 'volume')
        self.assertIsNone(ret)

    def test_ensure_share(self):
        share = self.share1
        gmgr1 = common.GlusterManager(self.glusterfs_target1, self._execute,
                                      None, None)
        gmgr1.set_vol_option = mock.Mock()
        self.mock_object(self._layout, '_share_manager',
                         mock.Mock(return_value=gmgr1))

        self._layout.ensure_share(self._context, share)

        self._layout._share_manager.assert_called_once_with(share)
        self.assertIn(self.glusterfs_target1, self._layout.gluster_used_vols)
        gmgr1.set_vol_option.assert_called_once_with(
            'user.manila-share', share['id'])

    @ddt.data({"voldict": {"host:/share2G": {"size": 2}}, "used_vols": set(),
               "size": 1, "expected": "host:/share2G"},
              {"voldict": {"host:/share2G": {"size": 2}}, "used_vols": set(),
               "size": 2, "expected": "host:/share2G"},
              {"voldict": {"host:/share2G": {"size": 2}}, "used_vols": set(),
               "size": None, "expected": "host:/share2G"},
              {"voldict": {"host:/share2G": {"size": 2},
                           "host:/share": {"size": None}},
               "used_vols": set(["host:/share2G"]), "size": 1,
               "expected": "host:/share"},
              {"voldict": {"host:/share2G": {"size": 2},
                           "host:/share": {"size": None}},
               "used_vols": set(["host:/share2G"]), "size": 2,
               "expected": "host:/share"},
              {"voldict": {"host:/share2G": {"size": 2},
               "host:/share": {"size": None}},
               "used_vols": set(["host:/share2G"]), "size": 3,
               "expected": "host:/share"},
              {"voldict": {"host:/share2G": {"size": 2},
                           "host:/share": {"size": None}},
               "used_vols": set(["host:/share2G"]), "size": None,
               "expected": "host:/share"},
              {"voldict": {"host:/share": {}}, "used_vols": set(), "size": 1,
               "expected": "host:/share"},
              {"voldict": {"host:/share": {}}, "used_vols": set(),
               "size": None, "expected": "host:/share"})
    @ddt.unpack
    def test_pop_gluster_vol(self, voldict, used_vols, size, expected):
        gmgr = common.GlusterManager
        gmgr1 = gmgr(expected, self._execute, None, None)
        self._layout._fetch_gluster_volumes = mock.Mock(return_value=voldict)
        self._layout.gluster_used_vols = used_vols
        self._layout._glustermanager = mock.Mock(return_value=gmgr1)
        self._layout.volume_pattern_keys = list(voldict.values())[0].keys()

        result = self._layout._pop_gluster_vol(size=size)

        self.assertEqual(expected, result)
        self.assertIn(result, used_vols)
        self._layout._fetch_gluster_volumes.assert_called_once_with()
        self._layout._glustermanager.assert_called_once_with(result)

    @ddt.data({"voldict": {"share2G": {"size": 2}},
               "used_vols": set(), "size": 3},
              {"voldict": {"share2G": {"size": 2}},
               "used_vols": set(["share2G"]), "size": None})
    @ddt.unpack
    def test_pop_gluster_vol_excp(self, voldict, used_vols, size):
        self._layout._fetch_gluster_volumes = mock.Mock(return_value=voldict)
        self._layout.gluster_used_vols = used_vols
        self._layout.volume_pattern_keys = list(voldict.values())[0].keys()

        self.assertRaises(exception.GlusterfsException,
                          self._layout._pop_gluster_vol, size=size)

        self._layout._fetch_gluster_volumes.assert_called_once_with()
        self.assertFalse(
            self.fake_driver._setup_via_manager.called)

    def test_push_gluster_vol(self):
        self._layout.gluster_used_vols = set([
            self.glusterfs_target1, self.glusterfs_target2])

        self._layout._push_gluster_vol(self.glusterfs_target2)

        self.assertEqual(1, len(self._layout.gluster_used_vols))
        self.assertFalse(
            self.glusterfs_target2 in self._layout.gluster_used_vols)

    def test_push_gluster_vol_excp(self):
        self._layout.gluster_used_vols = set([self.glusterfs_target1])
        self._layout.gluster_unused_vols_dict = {}

        self.assertRaises(exception.GlusterfsException,
                          self._layout._push_gluster_vol,
                          self.glusterfs_target2)

    @ddt.data({'vers_minor': '6',
               'cmd': ['find', '/tmp/tmpKGHKJ', '-mindepth', '1',
                       '-delete']},
              {'vers_minor': '7',
               'cmd': ['find', '/tmp/tmpKGHKJ', '-mindepth', '1', '!',
                       '-path', '/tmp/tmpKGHKJ/.trashcan', '!', '-path',
                       '/tmp/tmpKGHKJ/.trashcan/internal_op', '-delete']})
    @ddt.unpack
    def test_wipe_gluster_vol(self, vers_minor, cmd):
        tmpdir = '/tmp/tmpKGHKJ'
        gmgr = common.GlusterManager
        gmgr1 = gmgr(self.glusterfs_target1, self._execute, None, None)
        self._layout.glusterfs_versions = {
            self.glusterfs_server1: ('3', vers_minor)}

        self.mock_object(tempfile, 'mkdtemp',
                         mock.Mock(return_value=tmpdir))
        self.mock_object(self.fake_driver, '_execute', mock.Mock())
        self.mock_object(common, '_mount_gluster_vol', mock.Mock())
        self.mock_object(common, '_umount_gluster_vol', mock.Mock())
        self.mock_object(shutil, 'rmtree', mock.Mock())

        self._layout._wipe_gluster_vol(gmgr1)

        tempfile.mkdtemp.assert_called_once_with()
        common._mount_gluster_vol.assert_called_once_with(
            self.fake_driver._execute, gmgr1.export,
            tmpdir)
        kwargs = {'run_as_root': True}
        self.fake_driver._execute.assert_called_once_with(
            *cmd, **kwargs)
        common._umount_gluster_vol.assert_called_once_with(
            self.fake_driver._execute, tmpdir)
        kwargs = {'ignore_errors': True}
        shutil.rmtree.assert_called_once_with(tmpdir,
                                              **kwargs)

    def test_wipe_gluster_vol_mount_fail(self):
        tmpdir = '/tmp/tmpKGHKJ'
        gmgr = common.GlusterManager
        gmgr1 = gmgr(self.glusterfs_target1, self._execute, None, None)
        self._layout.glusterfs_versions = {
            self.glusterfs_server1: ('3', '6')}
        self.mock_object(tempfile, 'mkdtemp',
                         mock.Mock(return_value=tmpdir))
        self.mock_object(self.fake_driver, '_execute', mock.Mock())
        self.mock_object(common, '_mount_gluster_vol',
                         mock.Mock(side_effect=exception.GlusterfsException))
        self.mock_object(common, '_umount_gluster_vol', mock.Mock())
        self.mock_object(shutil, 'rmtree', mock.Mock())

        self.assertRaises(exception.GlusterfsException,
                          self._layout._wipe_gluster_vol,
                          gmgr1)

        tempfile.mkdtemp.assert_called_once_with()
        common._mount_gluster_vol.assert_called_once_with(
            self.fake_driver._execute, gmgr1.export,
            tmpdir)
        self.assertFalse(self.fake_driver._execute.called)
        self.assertFalse(common._umount_gluster_vol.called)
        kwargs = {'ignore_errors': True}
        shutil.rmtree.assert_called_once_with(tmpdir,
                                              **kwargs)

    def test_wipe_gluster_vol_error_wiping_gluster_vol(self):
        tmpdir = '/tmp/tmpKGHKJ'
        gmgr = common.GlusterManager
        gmgr1 = gmgr(self.glusterfs_target1, self._execute, None, None)
        self._layout.glusterfs_versions = {
            self.glusterfs_server1: ('3', '6')}
        cmd = ['find', '/tmp/tmpKGHKJ', '-mindepth', '1', '-delete']
        self.mock_object(tempfile, 'mkdtemp',
                         mock.Mock(return_value=tmpdir))
        self.mock_object(
            self.fake_driver, '_execute',
            mock.Mock(side_effect=exception.ProcessExecutionError))
        self.mock_object(common, '_mount_gluster_vol', mock.Mock())
        self.mock_object(common, '_umount_gluster_vol', mock.Mock())
        self.mock_object(shutil, 'rmtree', mock.Mock())

        self.assertRaises(exception.GlusterfsException,
                          self._layout._wipe_gluster_vol,
                          gmgr1)

        tempfile.mkdtemp.assert_called_once_with()
        common._mount_gluster_vol.assert_called_once_with(
            self.fake_driver._execute, gmgr1.export,
            tmpdir)
        kwargs = {'run_as_root': True}
        self.fake_driver._execute.assert_called_once_with(
            *cmd, **kwargs)
        common._umount_gluster_vol.assert_called_once_with(
            self.fake_driver._execute, tmpdir)
        kwargs = {'ignore_errors': True}
        shutil.rmtree.assert_called_once_with(tmpdir,
                                              **kwargs)

    def test_create_share(self):
        self._layout._pop_gluster_vol = mock.Mock(
            return_value=self.glusterfs_target1)
        gmgr1 = common.GlusterManager(self.glusterfs_target1)
        gmgr1.set_vol_option = mock.Mock()
        self.mock_object(self._layout, '_glustermanager',
                         mock.Mock(return_value=gmgr1))
        self.mock_object(self.fake_driver, '_setup_via_manager',
                         mock.Mock(return_value='host1:/gv1'))

        share = new_share()
        exp_locn = self._layout.create_share(self._context, share)

        self._layout._pop_gluster_vol.assert_called_once_with(share['size'])
        self.fake_driver._setup_via_manager.assert_called_once_with(
            {'manager': gmgr1, 'share': share})
        self._layout.private_storage.update.assert_called_once_with(
            share['id'], {'volume': self.glusterfs_target1})
        gmgr1.set_vol_option.assert_called_once_with(
            'user.manila-share', share['id'])
        self.assertEqual('host1:/gv1', exp_locn)

    def test_create_share_error(self):
        self._layout._pop_gluster_vol = mock.Mock(
            side_effect=exception.GlusterfsException)

        share = new_share()
        self.assertRaises(exception.GlusterfsException,
                          self._layout.create_share, self._context, share)

        self._layout._pop_gluster_vol.assert_called_once_with(
            share['size'])

    @ddt.data(None, '', 'Eeyore')
    def test_delete_share(self, clone_of):
        self._layout._push_gluster_vol = mock.Mock()
        self._layout._wipe_gluster_vol = mock.Mock()
        gmgr = common.GlusterManager
        gmgr1 = gmgr(self.glusterfs_target1, self._execute, None, None)
        gmgr1.set_vol_option = mock.Mock()
        gmgr1.get_vol_option = mock.Mock(return_value=clone_of)
        new_vol_addr = self.glusterfs_target1
        self.mock_object(self._layout, '_glustermanager',
                         mock.Mock(return_value=gmgr1))
        self._layout.gluster_used_vols = set([self.glusterfs_target1])

        self._layout.delete_share(self._context, self.share1)

        gmgr1.get_vol_option.assert_called_once_with(
            'user.manila-cloned-from')
        self._layout._wipe_gluster_vol.assert_called_once_with(gmgr1)
        self.assertIn(new_vol_addr, self._layout.gluster_used_vols)
        self._layout._push_gluster_vol.assert_called_once_with(
            self.glusterfs_target1)
        self._layout.private_storage.delete.assert_called_once_with(
            self.share1['id'])
        gmgr1.set_vol_option.assert_has_calls([
            mock.call('user.manila-share', 'NONE'),
            mock.call('nfs.disable', 'on')
        ])

    def test_delete_share_clone(self):
        self._layout._push_gluster_vol = mock.Mock()
        self._layout._wipe_gluster_vol = mock.Mock()
        gmgr = common.GlusterManager
        gmgr1 = gmgr(self.glusterfs_target1, self._execute, None, None)
        gmgr1.gluster_call = mock.Mock()
        gmgr1.get_vol_option = mock.Mock(return_value=FAKE_UUID1)
        self.mock_object(self._layout, '_glustermanager',
                         mock.Mock(return_value=gmgr1))
        self._layout.gluster_used_vols = set([self.glusterfs_target1])

        self._layout.delete_share(self._context, self.share1)

        gmgr1.get_vol_option.assert_called_once_with(
            'user.manila-cloned-from')
        self.assertFalse(self._layout._wipe_gluster_vol.called)
        self._layout._push_gluster_vol.assert_called_once_with(
            self.glusterfs_target1)
        self._layout.private_storage.delete.assert_called_once_with(
            self.share1['id'])
        gmgr1.gluster_call.assert_called_once_with(
            'volume', 'delete', 'gv1')

    def test_delete_share_error(self):
        self._layout._wipe_gluster_vol = mock.Mock()
        self._layout._wipe_gluster_vol.side_effect = (
            exception.GlusterfsException)
        self._layout._push_gluster_vol = mock.Mock()
        gmgr = common.GlusterManager
        gmgr1 = gmgr(self.glusterfs_target1, self._execute, None, None)
        gmgr1.get_vol_option = mock.Mock(return_value=None)
        self.mock_object(self._layout, '_glustermanager',
                         mock.Mock(return_value=gmgr1))
        self._layout.gluster_used_vols = set([self.glusterfs_target1])

        self.assertRaises(exception.GlusterfsException,
                          self._layout.delete_share, self._context,
                          self.share1)

        self._layout._wipe_gluster_vol.assert_called_once_with(gmgr1)
        self.assertFalse(self._layout._push_gluster_vol.called)

    def test_delete_share_missing_record(self):
        self.mock_object(self._layout, '_share_manager',
                         mock.Mock(return_value=None))

        self._layout.delete_share(self._context, self.share1)

        self._layout._share_manager.assert_called_once_with(self.share1)

    def test_create_snapshot(self):
        self._layout.gluster_nosnap_vols_dict = {}
        self._layout.glusterfs_versions = {self.glusterfs_server1: ('3', '6')}
        gmgr = common.GlusterManager
        gmgr1 = gmgr(self.glusterfs_target1, self._execute, None, None)
        self._layout.gluster_used_vols = set([self.glusterfs_target1])
        self.mock_object(gmgr1, 'gluster_call',
                         mock.Mock(
                             side_effect=(glusterXMLOut(ret=0, errno=0),)))
        self.mock_object(self._layout, '_glustermanager',
                         mock.Mock(return_value=gmgr1))

        snapshot = {
            'id': 'fake_snap_id',
            'share_id': self.share1['id'],
            'share': self.share1
        }
        ret = self._layout.create_snapshot(self._context, snapshot)

        self.assertIsNone(ret)
        args = ('--xml', 'snapshot', 'create', 'manila-fake_snap_id',
                gmgr1.volume)
        gmgr1.gluster_call.assert_called_once_with(*args, log=mock.ANY)

    @ddt.data({'side_effect': (glusterXMLOut(ret=-1, errno=2),),
               '_exception': exception.GlusterfsException},
              {'side_effect': (('', ''),),
               '_exception': exception.GlusterfsException})
    @ddt.unpack
    def test_create_snapshot_error(self, side_effect, _exception):
        self._layout.gluster_nosnap_vols_dict = {}
        self._layout.glusterfs_versions = {self.glusterfs_server1: ('3', '6')}
        gmgr = common.GlusterManager
        gmgr1 = gmgr(self.glusterfs_target1, self._execute, None, None)
        self._layout.gluster_used_vols = set([self.glusterfs_target1])
        self.mock_object(gmgr1, 'gluster_call',
                         mock.Mock(side_effect=side_effect))
        self.mock_object(self._layout, '_glustermanager',
                         mock.Mock(return_value=gmgr1))

        snapshot = {
            'id': 'fake_snap_id',
            'share_id': self.share1['id'],
            'share': self.share1
        }
        self.assertRaises(_exception, self._layout.create_snapshot,
                          self._context, snapshot)

        args = ('--xml', 'snapshot', 'create', 'manila-fake_snap_id',
                gmgr1.volume)
        gmgr1.gluster_call.assert_called_once_with(*args, log=mock.ANY)

    @ddt.data({"vers_minor": '6', "exctype": exception.GlusterfsException},
              {"vers_minor": '7',
               "exctype": exception.ShareSnapshotNotSupported})
    @ddt.unpack
    def test_create_snapshot_no_snap(self, vers_minor, exctype):
        self._layout.gluster_nosnap_vols_dict = {}
        self._layout.glusterfs_versions = {
            self.glusterfs_server1: ('3', vers_minor)}
        gmgr = common.GlusterManager
        gmgr1 = gmgr(self.glusterfs_target1, self._execute, None, None)
        self._layout.gluster_used_vols = set([self.glusterfs_target1])
        self.mock_object(gmgr1, 'gluster_call',
                         mock.Mock(
                             side_effect=(glusterXMLOut(ret=-1, errno=0),)))
        self.mock_object(self._layout, '_glustermanager',
                         mock.Mock(return_value=gmgr1))

        snapshot = {
            'id': 'fake_snap_id',
            'share_id': self.share1['id'],
            'share': self.share1
        }
        self.assertRaises(exctype, self._layout.create_snapshot, self._context,
                          snapshot)

        args = ('--xml', 'snapshot', 'create', 'manila-fake_snap_id',
                gmgr1.volume)
        gmgr1.gluster_call.assert_called_once_with(*args, log=mock.ANY)

    @ddt.data({"vers_minor": '6', "exctype": exception.GlusterfsException},
              {"vers_minor": '7',
               "exctype": exception.ShareSnapshotNotSupported})
    @ddt.unpack
    def test_create_snapshot_no_snap_cached(self, vers_minor, exctype):
        self._layout.gluster_nosnap_vols_dict = {
            self.glusterfs_target1: 'fake error'}
        self._layout.glusterfs_versions = {
            self.glusterfs_server1: ('3', vers_minor)}
        self._layout.gluster_used_vols = set([self.glusterfs_target1])
        gmgr = common.GlusterManager
        gmgr1 = gmgr(self.glusterfs_target1, self._execute, None, None)
        self.mock_object(self._layout, '_share_manager',
                         mock.Mock(return_value=gmgr1))

        snapshot = {
            'id': 'fake_snap_id',
            'share_id': self.share1['id'],
            'share': self.share1
        }
        self.assertRaises(exctype, self._layout.create_snapshot, self._context,
                          snapshot)

    def test_find_actual_backend_snapshot_name(self):
        gmgr = common.GlusterManager
        gmgr1 = gmgr(self.share1['export_location'], self._execute, None, None)
        self.mock_object(gmgr1, 'gluster_call',
                         mock.Mock(return_value=('fake_snap_id_xyz', '')))

        snapshot = {
            'id': 'fake_snap_id',
            'share_id': self.share1['id'],
            'share': self.share1
        }
        ret = self._layout._find_actual_backend_snapshot_name(gmgr1, snapshot)

        args = ('snapshot', 'list', gmgr1.volume, '--mode=script')
        gmgr1.gluster_call.assert_called_once_with(*args, log=mock.ANY)
        self.assertEqual('fake_snap_id_xyz', ret)

    @ddt.data('this is too bad', 'fake_snap_id_xyx\nfake_snap_id_pqr')
    def test_find_actual_backend_snapshot_name_bad_snap_list(self, snaplist):
        gmgr = common.GlusterManager
        gmgr1 = gmgr(self.share1['export_location'], self._execute, None, None)
        self.mock_object(gmgr1, 'gluster_call',
                         mock.Mock(return_value=(snaplist, '')))

        snapshot = {
            'id': 'fake_snap_id',
            'share_id': self.share1['id'],
            'share': self.share1
        }
        self.assertRaises(exception.GlusterfsException,
                          self._layout._find_actual_backend_snapshot_name,
                          gmgr1, snapshot)

        args = ('snapshot', 'list', gmgr1.volume, '--mode=script')
        gmgr1.gluster_call.assert_called_once_with(*args, log=mock.ANY)

    @ddt.data({'glusterfs_target': 'root@host1:/gv1',
               'glusterfs_server': 'root@host1'},
              {'glusterfs_target': 'host1:/gv1',
               'glusterfs_server': 'host1'})
    @ddt.unpack
    def test_create_share_from_snapshot(self, glusterfs_target,
                                        glusterfs_server):
        share = new_share()
        snapshot = {
            'id': 'fake_snap_id',
            'share_instance': new_share(export_location=glusterfs_target),
            'share_id': 'fake_share_id',
        }
        volume = ''.join(['manila-', share['id']])
        new_vol_addr = ':/'.join([glusterfs_server, volume])
        gmgr = common.GlusterManager
        old_gmgr = gmgr(glusterfs_target, self._execute, None, None)
        new_gmgr = gmgr(new_vol_addr, self._execute, None, None)
        self._layout.gluster_used_vols = set([glusterfs_target])
        self._layout.glusterfs_versions = {glusterfs_server: ('3', '7')}
        self.mock_object(old_gmgr, 'gluster_call',
                         mock.Mock(side_effect=[
                             ('', ''), ('', ''), ('', ''), ('', '')]))
        self.mock_object(new_gmgr, 'gluster_call',
                         mock.Mock(side_effect=[('', ''), ('', '')]))
        self.mock_object(new_gmgr, 'get_vol_option',
                         mock.Mock())
        new_gmgr.get_vol_option.return_value = (
            'glusterfs-server-1,client')
        self.mock_object(self._layout, '_find_actual_backend_snapshot_name',
                         mock.Mock(return_value='fake_snap_id_xyz'))
        self.mock_object(self._layout, '_share_manager',
                         mock.Mock(return_value=old_gmgr))
        self.mock_object(self._layout, '_glustermanager',
                         mock.Mock(return_value=new_gmgr))
        self.mock_object(self.fake_driver, '_setup_via_manager',
                         mock.Mock(return_value='host1:/gv1'))

        ret = self._layout.create_share_from_snapshot(
            self._context, share, snapshot, None)

        (self._layout._find_actual_backend_snapshot_name.
            assert_called_once_with(old_gmgr, snapshot))
        args = (('snapshot', 'activate', 'fake_snap_id_xyz',
                 'force', '--mode=script'),
                ('snapshot', 'clone', volume, 'fake_snap_id_xyz'),
                ('volume', 'start', volume))
        old_gmgr.gluster_call.assert_has_calls(
            [mock.call(*a, log=mock.ANY) for a in args])
        args = (('volume', 'set', volume, 'user.manila-share', share['id']),
                ('volume', 'set', volume, 'user.manila-cloned-from',
                 snapshot['share_id']))
        new_gmgr.gluster_call.assert_has_calls(
            [mock.call(*a, log=mock.ANY) for a in args], any_order=True)
        self._layout._share_manager.assert_called_once_with(
            snapshot['share_instance'])
        self._layout._glustermanager.assert_called_once_with(
            gmgr.parse(new_vol_addr))
        self._layout.driver._setup_via_manager.assert_called_once_with(
            {'manager': new_gmgr, 'share': share},
            {'manager': old_gmgr, 'share': snapshot['share_instance']})
        self._layout.private_storage.update.assert_called_once_with(
            share['id'], {'volume': new_vol_addr})
        self.assertIn(
            new_vol_addr,
            self._layout.gluster_used_vols)
        self.assertEqual(['host1:/gv1'], ret)

    def test_create_share_from_snapshot_error_unsupported_gluster_version(
            self):
        glusterfs_target = 'root@host1:/gv1'
        glusterfs_server = 'root@host1'
        share = new_share()
        volume = ''.join(['manila-', share['id']])
        new_vol_addr = ':/'.join([glusterfs_server, volume])
        gmgr = common.GlusterManager
        old_gmgr = gmgr(glusterfs_target, self._execute, None, None)
        new_gmgr = gmgr(new_vol_addr, self._execute, None, None)
        self._layout.gluster_used_vols_dict = {glusterfs_target: old_gmgr}
        self._layout.glusterfs_versions = {glusterfs_server: ('3', '6')}
        self.mock_object(
            old_gmgr, 'gluster_call',
            mock.Mock(side_effect=[('', ''), ('', '')]))
        self.mock_object(new_gmgr, 'get_vol_option',
                         mock.Mock())
        new_gmgr.get_vol_option.return_value = (
            'glusterfs-server-1,client')
        self.mock_object(self._layout, '_find_actual_backend_snapshot_name',
                         mock.Mock(return_value='fake_snap_id_xyz'))
        self.mock_object(self._layout, '_share_manager',
                         mock.Mock(return_value=old_gmgr))
        self.mock_object(self._layout, '_glustermanager',
                         mock.Mock(return_value=new_gmgr))

        snapshot = {
            'id': 'fake_snap_id',
            'share_instance': new_share(export_location=glusterfs_target)
        }
        self.assertRaises(exception.GlusterfsException,
                          self._layout.create_share_from_snapshot,
                          self._context, share, snapshot)

        self.assertFalse(
            self._layout._find_actual_backend_snapshot_name.called)
        self.assertFalse(old_gmgr.gluster_call.called)
        self._layout._share_manager.assert_called_once_with(
            snapshot['share_instance'])
        self.assertFalse(self._layout._glustermanager.called)
        self.assertFalse(new_gmgr.get_vol_option.called)
        self.assertFalse(new_gmgr.gluster_call.called)
        self.assertNotIn(new_vol_addr,
                         self._layout.glusterfs_versions.keys())

    def test_delete_snapshot(self):
        self._layout.gluster_nosnap_vols_dict = {}
        gmgr = common.GlusterManager
        gmgr1 = gmgr(self.share1['export_location'], self._execute, None, None)
        self._layout.gluster_used_vols = set([self.glusterfs_target1])
        self.mock_object(self._layout, '_find_actual_backend_snapshot_name',
                         mock.Mock(return_value='fake_snap_id_xyz'))
        self.mock_object(
            gmgr1, 'gluster_call',
            mock.Mock(return_value=glusterXMLOut(ret=0, errno=0)))
        self.mock_object(self._layout, '_glustermanager',
                         mock.Mock(return_value=gmgr1))
        snapshot = {
            'id': 'fake_snap_id',
            'share_id': self.share1['id'],
            'share': self.share1
        }
        ret = self._layout.delete_snapshot(self._context, snapshot)

        self.assertIsNone(ret)
        args = ('--xml', 'snapshot', 'delete', 'fake_snap_id_xyz',
                '--mode=script')
        gmgr1.gluster_call.assert_called_once_with(*args, log=mock.ANY)
        (self._layout._find_actual_backend_snapshot_name.
            assert_called_once_with(gmgr1, snapshot))

    @ddt.data({'side_effect': (glusterXMLOut(ret=-1, errno=0),),
               '_exception': exception.GlusterfsException},
              {'side_effect': (('', ''),),
               '_exception': exception.GlusterfsException})
    @ddt.unpack
    def test_delete_snapshot_error(self, side_effect, _exception):
        self._layout.gluster_nosnap_vols_dict = {}
        gmgr = common.GlusterManager
        gmgr1 = gmgr(self.share1['export_location'], self._execute, None, None)
        self._layout.gluster_used_vols = set([self.glusterfs_target1])
        self.mock_object(self._layout, '_find_actual_backend_snapshot_name',
                         mock.Mock(return_value='fake_snap_id_xyz'))
        args = ('--xml', 'snapshot', 'delete', 'fake_snap_id_xyz',
                '--mode=script')
        self.mock_object(
            gmgr1, 'gluster_call',
            mock.Mock(side_effect=side_effect))
        self.mock_object(self._layout, '_glustermanager',
                         mock.Mock(return_value=gmgr1))

        snapshot = {
            'id': 'fake_snap_id',
            'share_id': self.share1['id'],
            'share': self.share1
        }
        self.assertRaises(_exception, self._layout.delete_snapshot,
                          self._context, snapshot)

        gmgr1.gluster_call.assert_called_once_with(*args, log=mock.ANY)
        (self._layout._find_actual_backend_snapshot_name.
            assert_called_once_with(gmgr1, snapshot))

    @ddt.data(
        ('manage_existing', ('share', 'driver_options'), {}),
        ('unmanage', ('share',), {}),
        ('extend_share', ('share', 'new_size'), {'share_server': None}),
        ('shrink_share', ('share', 'new_size'), {'share_server': None}))
    def test_nonimplemented_methods(self, method_invocation):
        method, args, kwargs = method_invocation
        self.assertRaises(NotImplementedError, getattr(self._layout, method),
                          *args, **kwargs)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Add task_state field for share servers table

Revision ID: 5aa813ae673d
Revises: e6d88547b381
Create Date: 2020-06-23 12:04:47.821793

"""

# revision identifiers, used by Alembic.
revision = '5aa813ae673d'
down_revision = 'e6d88547b381'

from alembic import op
from oslo_log import log
import sqlalchemy as sa


LOG = log.getLogger(__name__)


share_servers_fk_name = (
    "fk_share_servers_source_share_server_id")


def upgrade():

    try:
        op.add_column('share_servers', sa.Column(
            'task_state', sa.String(length=255), default=None))
        op.add_column(
            'share_servers', sa.Column(
                'source_share_server_id', sa.String(length=36),
                sa.ForeignKey('share_servers.id', name=share_servers_fk_name),
                default=None,
                nullable=True))

    except Exception:
        LOG.error("Column share_servers.task_state and/or "
                  "share_server.source_share_server_id not created!")
        raise


def downgrade():
    try:
        op.drop_column('share_servers', 'task_state')
        op.drop_constraint(share_servers_fk_name, 'share_servers',
                           type_='foreignkey')
        op.drop_column('share_servers', 'source_share_server_id')
    except Exception:
        LOG.error("Column share_servers.task_state and/or "
                  "share_servers.source_share_server_id not dropped!")
        raise



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""add_per_share_gigabytes_quota_class

Revision ID: 0c23aec99b74
Revises: 5aa813ae673d
Create Date: 2021-01-03 10:01:57.276225

"""

# revision identifiers, used by Alembic.
revision = '0c23aec99b74'
down_revision = '5aa813ae673d'

from alembic import op
from manila.db.migrations import utils
from oslo_log import log
from oslo_utils import timeutils
from sqlalchemy import MetaData

LOG = log.getLogger(__name__)


def upgrade():
    meta = MetaData()
    meta.bind = op.get_bind()
    connection = op.get_bind().connect()
    quota_classes_table = utils.load_table('quota_classes', connection)

    try:
        op.bulk_insert
        (quota_classes_table,
         [{'created_at': timeutils.utcnow(),
           'class_name': 'default',
           'resource': 'per_share_gigabytes',
           'hard_limit': -1,
           'deleted': False, }])
    except Exception:
        LOG.error("Default per_share_gigabytes row not inserted "
                  "into the quota_classes.")
        raise


def downgrade():
    """Don't delete the 'default' entries at downgrade time.

    We don't know if the user had default entries when we started.
    If they did, we wouldn't want to remove them.  So, the safest
    thing to do is just leave the 'default' entries at downgrade time.
    """
    pass



# Copyright 2019 Nexenta by DDN, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt
from oslo_utils import units

from manila import context
from manila.share.drivers.nexenta.ns5 import jsonrpc
from manila.share.drivers.nexenta.ns5 import nexenta_nas
from manila import test

RPC_PATH = 'manila.share.drivers.nexenta.ns5.jsonrpc'
DRV_PATH = 'manila.share.drivers.nexenta.ns5.nexenta_nas.NexentaNasDriver'
DRIVER_VERSION = '1.1'
SHARE = {'share_id': 'uuid', 'size': 1, 'share_proto': 'NFS'}
SHARE_PATH = 'pool1/nfs_share/share-uuid'
SHARE2 = {'share_id': 'uuid2', 'size': 2, 'share_proto': 'NFS'}
SHARE2_PATH = 'pool1/nfs_share/share-uuid2'
SNAPSHOT = {
    'snapshot_id': 'snap_id',
    'share': SHARE,
    'snapshot_path': '%s@%s' % (SHARE_PATH, 'snapshot-snap_id')}


@ddt.ddt
class TestNexentaNasDriver(test.TestCase):

    def setUp(self):
        def _safe_get(opt):
            return getattr(self.cfg, opt)
        self.cfg = mock.Mock()
        self.mock_object(
            self.cfg, 'safe_get', mock.Mock(side_effect=_safe_get))
        super(TestNexentaNasDriver, self).setUp()
        self.cfg.nexenta_nas_host = '1.1.1.1'
        self.cfg.nexenta_rest_addresses = ['2.2.2.2']
        self.ctx = context.get_admin_context()
        self.cfg.nexenta_rest_port = 8080
        self.cfg.nexenta_rest_protocol = 'auto'
        self.cfg.nexenta_pool = 'pool1'
        self.cfg.nexenta_dataset_record_size = 131072
        self.cfg.reserved_share_percentage = 0
        self.cfg.reserved_share_from_snapshot_percentage = 0
        self.cfg.nexenta_folder = 'nfs_share'
        self.cfg.nexenta_user = 'user'
        self.cfg.share_backend_name = 'NexentaStor5'
        self.cfg.nexenta_password = 'password'
        self.cfg.nexenta_thin_provisioning = False
        self.cfg.nexenta_mount_point_base = 'mnt'
        self.cfg.nexenta_rest_retry_count = 3
        self.cfg.nexenta_share_name_prefix = 'share-'
        self.cfg.max_over_subscription_ratio = 20.0
        self.cfg.enabled_share_protocols = 'NFS'
        self.cfg.nexenta_mount_point_base = '$state_path/mnt'
        self.cfg.nexenta_dataset_compression = 'on'
        self.cfg.network_config_group = 'DEFAULT'
        self.cfg.admin_network_config_group = (
            'fake_admin_network_config_group')
        self.cfg.driver_handles_share_servers = False
        self.cfg.safe_get = self.fake_safe_get
        self.nef_mock = mock.Mock()
        self.mock_object(jsonrpc, 'NefRequest')
        self.drv = nexenta_nas.NexentaNasDriver(configuration=self.cfg)
        self.drv.do_setup(self.ctx)

    def fake_safe_get(self, key):
        try:
            value = getattr(self.cfg, key)
        except AttributeError:
            value = None
        return value

    def test_backend_name(self):
        self.assertEqual('NexentaStor5', self.drv.share_backend_name)

    @mock.patch('%s._get_provisioned_capacity' % DRV_PATH)
    @mock.patch('manila.share.drivers.nexenta.ns5.'
                'jsonrpc.NefServices.get')
    @mock.patch('manila.share.drivers.nexenta.ns5.'
                'jsonrpc.NefFilesystems.set')
    @mock.patch('manila.share.drivers.nexenta.ns5.'
                'jsonrpc.NefFilesystems.get')
    def test_check_for_setup_error(self, get_filesystem, set_filesystem,
                                   get_service, prov_capacity):
        prov_capacity.return_value = 1
        get_filesystem.return_value = {
            'mountPoint': '/path/to/volume',
            'nonBlockingMandatoryMode': False,
            'smartCompression': False,
            'isMounted': True
        }
        get_service.return_value = {
            'state': 'online'
        }
        self.assertIsNone(self.drv.check_for_setup_error())
        get_filesystem.assert_called_with(self.drv.root_path)
        set_filesystem.assert_not_called()
        get_service.assert_called_with('nfs')
        get_filesystem.return_value = {
            'mountPoint': '/path/to/volume',
            'nonBlockingMandatoryMode': True,
            'smartCompression': True,
            'isMounted': True
        }
        set_filesystem.return_value = {}
        payload = {
            'nonBlockingMandatoryMode': False,
            'smartCompression': False
        }
        self.assertIsNone(self.drv.check_for_setup_error())
        get_filesystem.assert_called_with(self.drv.root_path)
        set_filesystem.assert_called_with(self.drv.root_path, payload)
        get_service.assert_called_with('nfs')
        get_filesystem.return_value = {
            'mountPoint': '/path/to/volume',
            'nonBlockingMandatoryMode': False,
            'smartCompression': True,
            'isMounted': True
        }
        payload = {
            'smartCompression': False
        }
        set_filesystem.return_value = {}
        self.assertIsNone(self.drv.check_for_setup_error())
        get_filesystem.assert_called_with(self.drv.root_path)
        set_filesystem.assert_called_with(self.drv.root_path, payload)
        get_service.assert_called_with('nfs')
        get_filesystem.return_value = {
            'mountPoint': '/path/to/volume',
            'nonBlockingMandatoryMode': True,
            'smartCompression': False,
            'isMounted': True
        }
        payload = {
            'nonBlockingMandatoryMode': False
        }
        set_filesystem.return_value = {}
        self.assertIsNone(self.drv.check_for_setup_error())
        get_filesystem.assert_called_with(self.drv.root_path)
        set_filesystem.assert_called_with(self.drv.root_path, payload)
        get_service.assert_called_with('nfs')
        get_filesystem.return_value = {
            'mountPoint': 'none',
            'nonBlockingMandatoryMode': False,
            'smartCompression': False,
            'isMounted': False
        }
        self.assertRaises(jsonrpc.NefException,
                          self.drv.check_for_setup_error)
        get_filesystem.return_value = {
            'mountPoint': '/path/to/volume',
            'nonBlockingMandatoryMode': False,
            'smartCompression': False,
            'isMounted': False
        }
        self.assertRaises(jsonrpc.NefException,
                          self.drv.check_for_setup_error)
        get_service.return_value = {
            'state': 'online'
        }
        self.assertRaises(jsonrpc.NefException,
                          self.drv.check_for_setup_error)

    @mock.patch('%s.NefFilesystems.get' % RPC_PATH)
    def test__get_provisioned_capacity(self, fs_get):
        fs_get.return_value = {
            'path': 'pool1/nfs_share/123',
            'referencedQuotaSize': 1 * units.Gi
        }

        self.drv._get_provisioned_capacity()

        self.assertEqual(1 * units.Gi, self.drv.provisioned_capacity)

    @mock.patch('%s._mount_filesystem' % DRV_PATH)
    @mock.patch('%s.NefFilesystems.create' % RPC_PATH)
    @mock.patch('%s.NefFilesystems.delete' % RPC_PATH)
    def test_create_share(self, delete_fs, create_fs, mount_fs):
        mount_path = '%s:/%s' % (self.cfg.nexenta_nas_host, SHARE_PATH)
        mount_fs.return_value = mount_path
        size = int(1 * units.Gi * 1.1)
        self.assertEqual(
            [{
                'path': mount_path,
                'id': 'share-uuid'
            }],
            self.drv.create_share(self.ctx, SHARE))

        payload = {
            'recordSize': 131072,
            'compressionMode': self.cfg.nexenta_dataset_compression,
            'path': SHARE_PATH,
            'referencedQuotaSize': size,
            'nonBlockingMandatoryMode': False,
            'referencedReservationSize': size
        }
        self.drv.nef.filesystems.create.assert_called_with(payload)

        mount_fs.side_effect = jsonrpc.NefException('some error')
        self.assertRaises(jsonrpc.NefException,
                          self.drv.create_share, self.ctx, SHARE)
        delete_payload = {'force': True}
        self.drv.nef.filesystems.delete.assert_called_with(
            SHARE_PATH, delete_payload)

    @mock.patch('%s.NefFilesystems.promote' % RPC_PATH)
    @mock.patch('%s.NefSnapshots.get' % RPC_PATH)
    @mock.patch('%s.NefSnapshots.list' % RPC_PATH)
    @mock.patch('%s.NefFilesystems.delete' % RPC_PATH)
    def test_delete_share(self, fs_delete, snap_list, snap_get, fs_promote):
        delete_payload = {'force': True, 'snapshots': True}
        snapshots_payload = {'parent': SHARE_PATH, 'fields': 'path'}
        clones_payload = {'fields': 'clones,creationTxg'}
        clone_path = '%s:/%s' % (self.cfg.nexenta_nas_host, 'path_to_fs')
        fs_delete.side_effect = [
            jsonrpc.NefException({
                'message': 'some_error',
                'code': 'EEXIST'}),
            None]
        snap_list.return_value = [{'path': '%s@snap1' % SHARE_PATH}]
        snap_get.return_value = {'clones': [clone_path], 'creationTxg': 1}
        self.assertIsNone(self.drv.delete_share(self.ctx, SHARE))
        fs_delete.assert_called_with(SHARE_PATH, delete_payload)
        fs_promote.assert_called_with(clone_path)
        snap_get.assert_called_with('%s@snap1' % SHARE_PATH, clones_payload)
        snap_list.assert_called_with(snapshots_payload)

    @mock.patch('%s.NefFilesystems.mount' % RPC_PATH)
    @mock.patch('%s.NefFilesystems.get' % RPC_PATH)
    def test_mount_filesystem(self, fs_get, fs_mount):
        mount_path = '%s:/%s' % (self.cfg.nexenta_nas_host, SHARE_PATH)
        fs_get.return_value = {
            'mountPoint': '/%s' % SHARE_PATH, 'isMounted': False}
        self.assertEqual(mount_path, self.drv._mount_filesystem(SHARE))
        self.drv.nef.filesystems.mount.assert_called_with(SHARE_PATH)

    @mock.patch('%s.NefHpr.activate' % RPC_PATH)
    @mock.patch('%s.NefFilesystems.mount' % RPC_PATH)
    @mock.patch('%s.NefFilesystems.get' % RPC_PATH)
    def test_mount_filesystem_with_activate(
            self, fs_get, fs_mount, hpr_activate):
        mount_path = '%s:/%s' % (self.cfg.nexenta_nas_host, SHARE_PATH)
        fs_get.side_effect = [
            {'mountPoint': 'none', 'isMounted': False},
            {'mountPoint': '/%s' % SHARE_PATH, 'isMounted': False}]
        self.assertEqual(mount_path, self.drv._mount_filesystem(SHARE))
        payload = {'datasetName': SHARE_PATH}
        self.drv.nef.hpr.activate.assert_called_once_with(payload)

    @mock.patch('%s.NefFilesystems.mount' % RPC_PATH)
    @mock.patch('%s.NefFilesystems.unmount' % RPC_PATH)
    def test_remount_filesystem(self, fs_unmount, fs_mount):
        self.drv._remount_filesystem(SHARE_PATH)
        fs_unmount.assert_called_once_with(SHARE_PATH)
        fs_mount.assert_called_once_with(SHARE_PATH)

    def parse_fqdn(self, fqdn):
        address_mask = fqdn.strip().split('/', 1)
        address = address_mask[0]
        ls = {"allow": True, "etype": "fqdn", "entity": address}
        if len(address_mask) == 2:
            ls['mask'] = address_mask[1]
            ls['etype'] = 'network'
        return ls

    @ddt.data({'key': 'value'}, {})
    @mock.patch('%s.NefNfs.list' % RPC_PATH)
    @mock.patch('%s.NefNfs.set' % RPC_PATH)
    @mock.patch('%s.NefFilesystems.acl' % RPC_PATH)
    def test_update_nfs_access(self, acl, nfs_set, nfs_list, list_data):
        security_contexts = {'securityModes': ['sys']}
        nfs_list.return_value = list_data
        rw_list = ['1.1.1.1/24', '2.2.2.2']
        ro_list = ['3.3.3.3', '4.4.4.4/30']
        security_contexts['readWriteList'] = []
        security_contexts['readOnlyList'] = []
        for fqdn in rw_list:
            ls = self.parse_fqdn(fqdn)
            if ls.get('mask'):
                ls['mask'] = int(ls['mask'])
            security_contexts['readWriteList'].append(ls)
        for fqdn in ro_list:
            ls = self.parse_fqdn(fqdn)
            if ls.get('mask'):
                ls['mask'] = int(ls['mask'])
            security_contexts['readOnlyList'].append(ls)

        self.assertIsNone(self.drv._update_nfs_access(SHARE, rw_list, ro_list))
        payload = {
            'flags': ['file_inherit', 'dir_inherit'],
            'permissions': ['full_set'],
            'principal': 'everyone@',
            'type': 'allow'
        }
        self.drv.nef.filesystems.acl.assert_called_with(SHARE_PATH, payload)
        payload = {'securityContexts': [security_contexts]}
        if list_data:
            self.drv.nef.nfs.set.assert_called_with(SHARE_PATH, payload)
        else:
            payload['filesystem'] = SHARE_PATH
            self.drv.nef.nfs.create.assert_called_with(payload)

    def test_update_nfs_access_bad_mask(self):
        security_contexts = {'securityModes': ['sys']}
        rw_list = ['1.1.1.1/24', '2.2.2.2/1a']
        ro_list = ['3.3.3.3', '4.4.4.4/30']
        security_contexts['readWriteList'] = []
        security_contexts['readOnlyList'] = []
        for fqdn in rw_list:
            security_contexts['readWriteList'].append(self.parse_fqdn(fqdn))
        for fqdn in ro_list:
            security_contexts['readOnlyList'].append(self.parse_fqdn(fqdn))

        self.assertRaises(ValueError, self.drv._update_nfs_access,
                          SHARE, rw_list, ro_list)

    @mock.patch('%s._update_nfs_access' % DRV_PATH)
    def test_update_access__ip_rw(self, update_nfs_access):
        access = {
            'access_type': 'ip',
            'access_to': '1.1.1.1',
            'access_level': 'rw',
            'access_id': 'fake_id'
        }

        self.assertEqual(
            {'fake_id': {'state': 'active'}},
            self.drv.update_access(
                self.ctx, SHARE, [access], None, None))
        self.drv._update_nfs_access.assert_called_with(SHARE, ['1.1.1.1'], [])

    @mock.patch('%s._update_nfs_access' % DRV_PATH)
    def test_update_access__ip_ro(self, update_nfs_access):
        access = {
            'access_type': 'ip',
            'access_to': '1.1.1.1',
            'access_level': 'ro',
            'access_id': 'fake_id'
        }

        expected = {'fake_id': {'state': 'active'}}
        self.assertEqual(
            expected, self.drv.update_access(
                self.ctx, SHARE, [access], None, None))
        self.drv._update_nfs_access.assert_called_with(SHARE, [], ['1.1.1.1'])

    @ddt.data('rw', 'ro')
    def test_update_access__not_ip(self, access_level):
        access = {
            'access_type': 'username',
            'access_to': 'some_user',
            'access_level': access_level,
            'access_id': 'fake_id'
        }
        expected = {'fake_id': {'state': 'error'}}
        self.assertEqual(expected, self.drv.update_access(
            self.ctx, SHARE, [access], None, None))

    @mock.patch('%s._get_capacity_info' % DRV_PATH)
    @mock.patch('manila.share.driver.ShareDriver._update_share_stats')
    def test_update_share_stats(self, super_stats, info):
        info.return_value = (100, 90, 10)
        stats = {
            'vendor_name': 'Nexenta',
            'storage_protocol': 'NFS',
            'nfs_mount_point_base': self.cfg.nexenta_mount_point_base,
            'create_share_from_snapshot_support': True,
            'revert_to_snapshot_support': True,
            'snapshot_support': True,
            'driver_version': DRIVER_VERSION,
            'share_backend_name': self.cfg.share_backend_name,
            'pools': [{
                'compression': True,
                'pool_name': 'pool1',
                'total_capacity_gb': 100,
                'free_capacity_gb': 90,
                'provisioned_capacity_gb': 0,
                'max_over_subscription_ratio': 20.0,
                'reserved_percentage': (
                    self.cfg.reserved_share_percentage),
                'reserved_snapshot_percentage': (
                    self.cfg.reserved_share_from_snapshot_percentage),
                'thin_provisioning': self.cfg.nexenta_thin_provisioning,
            }],
        }

        self.drv._update_share_stats()

        self.assertEqual(stats, self.drv._stats)

    def test_get_capacity_info(self):
        self.drv.nef.get.return_value = {
            'bytesAvailable': 9 * units.Gi, 'bytesUsed': 1 * units.Gi}

        self.assertEqual((10, 9, 1), self.drv._get_capacity_info())

    @mock.patch('%s._set_reservation' % DRV_PATH)
    @mock.patch('%s._set_quota' % DRV_PATH)
    @mock.patch('%s.NefFilesystems.rename' % RPC_PATH)
    @mock.patch('%s.NefFilesystems.get' % RPC_PATH)
    def test_manage_existing(self, fs_get, fs_rename, set_res, set_quota):
        fs_get.return_value = {'referencedQuotaSize': 1073741824}
        old_path = '%s:/%s' % (self.cfg.nexenta_nas_host, 'path_to_fs')
        new_path = '%s:/%s' % (self.cfg.nexenta_nas_host, SHARE_PATH)
        SHARE['export_locations'] = [{'path': old_path}]
        expected = {'size': 2, 'export_locations': [{
            'path': new_path
        }]}
        self.assertEqual(expected, self.drv.manage_existing(SHARE, None))
        fs_rename.assert_called_with('path_to_fs', {'newPath': SHARE_PATH})
        set_res.assert_called_with(SHARE, 2)
        set_quota.assert_called_with(SHARE, 2)

    @mock.patch('%s.NefSnapshots.create' % RPC_PATH)
    def test_create_snapshot(self, snap_create):
        self.assertIsNone(self.drv.create_snapshot(self.ctx, SNAPSHOT))
        snap_create.assert_called_once_with({
            'path': SNAPSHOT['snapshot_path']})

    @mock.patch('%s.NefSnapshots.delete' % RPC_PATH)
    def test_delete_snapshot(self, snap_delete):
        self.assertIsNone(self.drv.delete_snapshot(self.ctx, SNAPSHOT))
        payload = {'defer': True}
        snap_delete.assert_called_once_with(
            SNAPSHOT['snapshot_path'], payload)

    @mock.patch('%s._mount_filesystem' % DRV_PATH)
    @mock.patch('%s._remount_filesystem' % DRV_PATH)
    @mock.patch('%s.NefFilesystems.delete' % RPC_PATH)
    @mock.patch('%s.NefSnapshots.clone' % RPC_PATH)
    def test_create_share_from_snapshot(
            self, snap_clone, fs_delete, remount_fs, mount_fs):
        mount_fs.return_value = 'mount_path'
        location = {
            'path': 'mount_path',
            'id': 'share-uuid2'
        }
        self.assertEqual([location], self.drv.create_share_from_snapshot(
            self.ctx, SHARE2, SNAPSHOT))

        size = int(SHARE2['size'] * units.Gi * 1.1)
        payload = {
            'targetPath': SHARE2_PATH,
            'referencedQuotaSize': size,
            'recordSize': self.cfg.nexenta_dataset_record_size,
            'compressionMode': self.cfg.nexenta_dataset_compression,
            'nonBlockingMandatoryMode': False,
            'referencedReservationSize': size
        }
        snap_clone.assert_called_once_with(SNAPSHOT['snapshot_path'], payload)

    @mock.patch('%s._mount_filesystem' % DRV_PATH)
    @mock.patch('%s._remount_filesystem' % DRV_PATH)
    @mock.patch('%s.NefFilesystems.delete' % RPC_PATH)
    @mock.patch('%s.NefSnapshots.clone' % RPC_PATH)
    def test_create_share_from_snapshot_error(
            self, snap_clone, fs_delete, remount_fs, mount_fs):
        fs_delete.side_effect = jsonrpc.NefException('delete error')
        mount_fs.side_effect = jsonrpc.NefException('create error')
        self.assertRaises(
            jsonrpc.NefException,
            self.drv.create_share_from_snapshot, self.ctx, SHARE2, SNAPSHOT)

        size = int(SHARE2['size'] * units.Gi * 1.1)
        payload = {
            'targetPath': SHARE2_PATH,
            'referencedQuotaSize': size,
            'recordSize': self.cfg.nexenta_dataset_record_size,
            'compressionMode': self.cfg.nexenta_dataset_compression,
            'nonBlockingMandatoryMode': False,
            'referencedReservationSize': size
        }
        snap_clone.assert_called_once_with(SNAPSHOT['snapshot_path'], payload)
        payload = {'force': True}
        fs_delete.assert_called_once_with(SHARE2_PATH, payload)

    @mock.patch('%s.NefFilesystems.rollback' % RPC_PATH)
    def test_revert_to_snapshot(self, fs_rollback):
        self.assertIsNone(self.drv.revert_to_snapshot(
            self.ctx, SNAPSHOT, [], []))
        payload = {'snapshot': 'snapshot-snap_id'}
        fs_rollback.assert_called_once_with(
            SHARE_PATH, payload)

    @mock.patch('%s._set_reservation' % DRV_PATH)
    @mock.patch('%s._set_quota' % DRV_PATH)
    def test_extend_share(self, set_quota, set_reservation):
        self.assertIsNone(self.drv.extend_share(
            SHARE, 2))
        set_quota.assert_called_once_with(
            SHARE, 2)
        set_reservation.assert_called_once_with(
            SHARE, 2)

    @mock.patch('%s.NefFilesystems.get' % RPC_PATH)
    @mock.patch('%s._set_reservation' % DRV_PATH)
    @mock.patch('%s._set_quota' % DRV_PATH)
    def test_shrink_share(self, set_quota, set_reservation, fs_get):
        fs_get.return_value = {
            'bytesUsedBySelf': 0.5 * units.Gi
        }
        self.assertIsNone(self.drv.shrink_share(
            SHARE2, 1))
        set_quota.assert_called_once_with(
            SHARE2, 1)
        set_reservation.assert_called_once_with(
            SHARE2, 1)

    @mock.patch('%s.NefFilesystems.set' % RPC_PATH)
    def test_set_quota(self, fs_set):
        quota = int(2 * units.Gi * 1.1)
        payload = {'referencedQuotaSize': quota}
        self.assertIsNone(self.drv._set_quota(
            SHARE, 2))
        fs_set.assert_called_once_with(SHARE_PATH, payload)

    @mock.patch('%s.NefFilesystems.set' % RPC_PATH)
    def test_set_reservation(self, fs_set):
        reservation = int(2 * units.Gi * 1.1)
        payload = {'referencedReservationSize': reservation}
        self.assertIsNone(self.drv._set_reservation(
            SHARE, 2))
        fs_set.assert_called_once_with(SHARE_PATH, payload)



# Copyright (c) 2021 SAP.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log

from manila import exception
from manila.scheduler.filters import base_host
from manila.share import api

LOG = log.getLogger(__name__)

AFFINITY_FILTER = 'same_host'
ANTI_AFFINITY_FILTER = 'different_host'


class AffinityBaseFilter(base_host.BaseHostFilter):
    """Base class of affinity filters"""
    _filter_type = None

    def __init__(self):
        self.share_api = api.API()

    def filter_all(self, filter_obj_list, filter_properties):
        # _filter_type should be defined in subclass
        if self._filter_type is None:
            raise AffinityFilterTypeNotSetError

        try:
            filter_properties = self._validate(filter_properties)
        except SchedulerHintsNotSet:
            # AffinityFilter/AntiAffinityFilter is skipped if corresponding
            # hint is not set. If the "scheduler_hints" is not set, both
            # filters are skipped.
            return filter_obj_list
        except (exception.InvalidUUID,
                exception.ShareNotFound,
                exception.ShareInstanceNotFound) as e:
            # Stop scheduling share when above errors are caught
            LOG.error('%(filter_name)s: %(error)s', {
                'filter_name': self.__class__.__name__,
                'error': e})
            return None
        else:
            # Return list of hosts which pass the function host_passes()
            # overriden in AffinityFilter and AntiAffinityFilter.
            return [obj for obj in filter_obj_list
                    if self._filter_one(obj, filter_properties)]

    def _validate(self, filter_properties):
        context = filter_properties['context']
        hints = filter_properties.get('scheduler_hints')

        if hints is None:
            raise SchedulerHintsNotSet
        else:
            share_uuids = hints.get(self._filter_type)
            if share_uuids is None:
                raise SchedulerHintsNotSet

        share_uuids = share_uuids.split(",")

        filter_properties['scheduler_hints'][self._filter_type] = []

        filtered_hosts = []
        for uuid in share_uuids:
            try:
                share = self.share_api.get(context, uuid)
            except exception.NotFound:
                raise exception.ShareNotFound(uuid)
            instances = share.get('instances')
            if len(instances) == 0:
                raise exception.ShareInstanceNotFound(share_instance_id=uuid)
            filtered_hosts.append(
                [instance.get('host') for instance in instances])

        if self._filter_type == AFFINITY_FILTER:
            filter_properties['scheduler_hints'][self._filter_type] = list(
                set.intersection(*map(set, filtered_hosts)))
        else:
            filter_properties['scheduler_hints'][self._filter_type] = list(
                set.union(*map(set, filtered_hosts)))

        return filter_properties


class AffinityFilter(AffinityBaseFilter):
    _filter_type = AFFINITY_FILTER

    def host_passes(self, host_state, filter_properties):
        allowed_hosts = \
            filter_properties['scheduler_hints'][self._filter_type]
        return host_state.host in allowed_hosts


class AntiAffinityFilter(AffinityBaseFilter):
    _filter_type = ANTI_AFFINITY_FILTER

    def host_passes(self, host_state, filter_properties):
        forbidden_hosts = \
            filter_properties['scheduler_hints'][self._filter_type]
        return host_state.host not in forbidden_hosts


class SchedulerHintsNotSet(Exception):
    pass


class AffinityFilterTypeNotSetError(Exception):
    pass



# Copyright (c) 2019 EMC Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
EMC specific NAS storage driver. This driver is a pluggable driver
that allows specific EMC NAS devices to be plugged-in as the underlying
backend. Use the Manila configuration variable "share_backend_name"
to specify, which backend plugins to use.
"""

from oslo_config import cfg
from oslo_log import log

from manila.share import driver
from manila.share.drivers.dell_emc import plugin_manager as manager

EMC_NAS_OPTS = [
    cfg.StrOpt('emc_nas_login',
               help='User name for the EMC server.'),
    cfg.StrOpt('emc_nas_password',
               help='Password for the EMC server.'),
    cfg.HostAddressOpt('emc_nas_server',
                       help='EMC server hostname or IP address.'),
    cfg.PortOpt('emc_nas_server_port',
                default=8080,
                help='Port number for the EMC server.'),
    cfg.BoolOpt('emc_nas_server_secure',
                default=True,
                help='Use secure connection to server.'),
    cfg.StrOpt('emc_share_backend',
               ignore_case=True,
               choices=['isilon', 'vnx', 'unity', 'vmax', 'powermax'],
               help='Share backend.'),
    cfg.StrOpt('emc_nas_root_dir',
               help='The root directory where shares will be located.'),
    cfg.BoolOpt('emc_ssl_cert_verify',
                default=True,
                help='If set to False the https client will not validate the '
                     'SSL certificate of the backend endpoint.'),
    cfg.StrOpt('emc_ssl_cert_path',
               help='Can be used to specify a non default path to a '
                    'CA_BUNDLE file or directory with certificates of trusted '
                    'CAs, which will be used to validate the backend.')
]

LOG = log.getLogger(__name__)

CONF = cfg.CONF
CONF.register_opts(EMC_NAS_OPTS)


class EMCShareDriver(driver.ShareDriver):
    """EMC specific NAS driver. Allows for NFS and CIFS NAS storage usage."""

    def __init__(self, *args, **kwargs):
        self.configuration = kwargs.get('configuration', None)
        if self.configuration:
            self.configuration.append_config_values(EMC_NAS_OPTS)
            self.backend_name = self.configuration.safe_get(
                'emc_share_backend')
        else:
            self.backend_name = CONF.emc_share_backend
        self.backend_name = self.backend_name or 'EMC_NAS_Storage'
        self.plugin_manager = manager.EMCPluginManager(
            namespace='manila.share.drivers.dell_emc.plugins')
        if self.backend_name == 'vmax':
            LOG.warning("Configuration option 'emc_share_backend=vmax' will "
                        "remain a valid option until the V release of "
                        "OpenStack. After that, only "
                        "'emc_share_backend=powermax' will be excepted.")
            self.backend_name = 'powermax'
        self.plugin = self.plugin_manager.load_plugin(
            self.backend_name,
            configuration=self.configuration)
        super(EMCShareDriver, self).__init__(
            self.plugin.driver_handles_share_servers, *args, **kwargs)

        self.dhss_mandatory_security_service_association = (
            self.plugin.dhss_mandatory_security_service_association)

        if hasattr(self.plugin, 'ipv6_implemented'):
            self.ipv6_implemented = self.plugin.ipv6_implemented

        if hasattr(self.plugin, 'revert_to_snap_support'):
            self.revert_to_snap_support = self.plugin.revert_to_snap_support
        else:
            self.revert_to_snap_support = False

        if hasattr(self.plugin, 'shrink_share_support'):
            self.shrink_share_support = self.plugin.shrink_share_support
        else:
            self.shrink_share_support = False

        if hasattr(self.plugin, 'manage_existing_support'):
            self.manage_existing_support = self.plugin.manage_existing_support
        else:
            self.manage_existing_support = False

        if hasattr(self.plugin, 'manage_existing_with_server_support'):
            self.manage_existing_with_server_support = (
                self.plugin.manage_existing_with_server_support)
        else:
            self.manage_existing_with_server_support = False

        if hasattr(self.plugin, 'manage_existing_snapshot_support'):
            self.manage_existing_snapshot_support = (
                self.plugin.manage_existing_snapshot_support)
        else:
            self.manage_existing_snapshot_support = False

        if hasattr(self.plugin, 'manage_snapshot_with_server_support'):
            self.manage_snapshot_with_server_support = (
                self.plugin.manage_snapshot_with_server_support)
        else:
            self.manage_snapshot_with_server_support = False

        if hasattr(self.plugin, 'manage_server_support'):
            self.manage_server_support = self.plugin.manage_server_support
        else:
            self.manage_server_support = False

        if hasattr(self.plugin, 'get_share_server_network_info_support'):
            self.get_share_server_network_info_support = (
                self.plugin.get_share_server_network_info_support)
        else:
            self.get_share_server_network_info_support = False

    def manage_existing(self, share, driver_options):
        """manage an existing share"""
        if self.manage_existing_support:
            return self.plugin.manage_existing(share, driver_options)
        else:
            return NotImplementedError()

    def manage_existing_with_server(self, share, driver_options,
                                    share_server=None):
        """manage an existing share"""
        if self.manage_existing_with_server_support:
            return self.plugin.manage_existing_with_server(
                share, driver_options, share_server)
        else:
            return NotImplementedError()

    def manage_existing_snapshot(self, snapshot, driver_options):
        """manage an existing share snapshot"""
        if self.manage_existing_snapshot_support:
            return self.plugin.manage_existing_snapshot(snapshot,
                                                        driver_options)
        else:
            return NotImplementedError()

    def manage_existing_snapshot_with_server(self, snapshot, driver_options,
                                             share_server=None):
        """manage an existing share snapshot"""
        if self.manage_snapshot_with_server_support:
            return self.plugin.manage_existing_snapshot_with_server(
                snapshot, driver_options, share_server=None)
        else:
            return NotImplementedError()

    def manage_server(self, context, share_server, identifier,
                      driver_options):
        if self.manage_server_support:
            return self.plugin.manage_server(context, share_server,
                                             identifier, driver_options)
        else:
            return NotImplementedError()

    def get_share_server_network_info(
            self, context, share_server, identifier, driver_options):
        if self.get_share_server_network_info_support:
            return self.plugin.get_share_server_network_info(
                context, share_server, identifier, driver_options)
        else:
            return NotImplementedError()

    def unmanage_server(self, server_details, security_services=None):
        LOG.info('Dell EMC driver will unmanage share server: %s out of '
                 'OpenStack.', server_details.get('server_id'))

    def unmanage(self, share):
        LOG.info('Dell EMC driver will unmanage share: %s out of '
                 'OpenStack.', share.get('id'))

    def unmanage_with_server(self, share, share_server=None):
        LOG.info('Dell EMC driver will unmanage share: %s out of '
                 'OpenStack.', share.get('id'))

    def unmanage_snapshot(self, snapshot):
        LOG.info('Dell EMC driver will unmanage snapshot: %s out of '
                 'OpenStack.', snapshot.get('id'))

    def unmanage_snapshot_with_server(self, snapshot, share_server=None):
        LOG.info('Dell EMC driver will unmanage snapshot: %s out of '
                 'OpenStack.', snapshot.get('id'))

    def create_share(self, context, share, share_server=None):
        """Is called to create share."""
        location = self.plugin.create_share(context, share, share_server)

        return location

    def create_share_from_snapshot(self, context, share, snapshot,
                                   share_server=None, parent_share=None):
        """Is called to create share from snapshot."""
        location = self.plugin.create_share_from_snapshot(
            context, share, snapshot, share_server)

        return location

    def extend_share(self, share, new_size, share_server=None):
        """Is called to extend share."""
        self.plugin.extend_share(share, new_size, share_server)

    def shrink_share(self, share, new_size, share_server=None):
        """Is called to shrink share."""
        if self.shrink_share_support:
            self.plugin.shrink_share(share, new_size, share_server)
        else:
            raise NotImplementedError()

    def create_snapshot(self, context, snapshot, share_server=None):
        """Is called to create snapshot."""
        return self.plugin.create_snapshot(context, snapshot, share_server)

    def delete_share(self, context, share, share_server=None):
        """Is called to remove share."""
        self.plugin.delete_share(context, share, share_server)

    def delete_snapshot(self, context, snapshot, share_server=None):
        """Is called to remove snapshot."""
        self.plugin.delete_snapshot(context, snapshot, share_server)

    def ensure_share(self, context, share, share_server=None):
        """Invoked to sure that share is exported."""
        self.plugin.ensure_share(context, share, share_server)

    def allow_access(self, context, share, access, share_server=None):
        """Allow access to the share."""
        self.plugin.allow_access(context, share, access, share_server)

    def deny_access(self, context, share, access, share_server=None):
        """Deny access to the share."""
        self.plugin.deny_access(context, share, access, share_server)

    def update_access(self, context, share, access_rules, add_rules,
                      delete_rules, share_server=None):
        """Update access to the share."""
        self.plugin.update_access(context, share, access_rules, add_rules,
                                  delete_rules, share_server)

    def check_for_setup_error(self):
        """Check for setup error."""
        self.plugin.check_for_setup_error()

    def do_setup(self, context):
        """Any initialization the share driver does while starting."""
        self.plugin.connect(self, context)

    def _update_share_stats(self):
        """Retrieve stats info from share."""

        backend_name = self.configuration.safe_get(
            'share_backend_name') or "EMC_NAS_Storage"
        data = dict(
            share_backend_name=backend_name,
            vendor_name='Dell EMC',
            storage_protocol='NFS_CIFS',
            snapshot_support=True,
            create_share_from_snapshot_support=True,
            revert_to_snapshot_support=self.revert_to_snap_support)
        self.plugin.update_share_stats(data)
        super(EMCShareDriver, self)._update_share_stats(data)

    def get_network_allocations_number(self):
        """Returns number of network allocations for creating VIFs."""
        return self.plugin.get_network_allocations_number()

    def _setup_server(self, network_info, metadata=None):
        """Set up and configures share server with given network parameters."""
        return self.plugin.setup_server(network_info, metadata)

    def _teardown_server(self, server_details, security_services=None):
        """Teardown share server."""
        return self.plugin.teardown_server(server_details, security_services)

    def get_configured_ip_versions(self):
        if self.ipv6_implemented:
            return [4, 6]
        else:
            return [4]

    def revert_to_snapshot(self, context, snapshot, share_access_rules,
                           snapshot_access_rules, share_server=None):
        if self.revert_to_snap_support:
            return self.plugin.revert_to_snapshot(context, snapshot,
                                                  share_access_rules,
                                                  snapshot_access_rules,
                                                  share_server)
        else:
            raise NotImplementedError()

    def get_default_filter_function(self):
        if hasattr(self.plugin, 'get_default_filter_function'):
            return self.plugin.get_default_filter_function()
        return None



# Copyright (c) 2015 Alex Meade
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Handles all requests relating to share groups.
"""

from oslo_config import cfg
from oslo_log import log
from oslo_utils import excutils
from oslo_utils import strutils
import six

from manila.api import common as api_common
from manila.common import constants
from manila.db import base
from manila import exception
from manila.i18n import _
from manila import quota
from manila.scheduler import rpcapi as scheduler_rpcapi
from manila import share
from manila.share import rpcapi as share_rpcapi
from manila.share import share_types

CONF = cfg.CONF
LOG = log.getLogger(__name__)
QUOTAS = quota.QUOTAS


class API(base.Base):
    """API for interacting with the share manager."""

    def __init__(self, db_driver=None):
        self.scheduler_rpcapi = scheduler_rpcapi.SchedulerAPI()
        self.share_rpcapi = share_rpcapi.ShareAPI()
        self.share_api = share.API()
        super(API, self).__init__(db_driver)

    def create(self, context, name=None, description=None,
               share_type_ids=None, source_share_group_snapshot_id=None,
               share_network_id=None, share_group_type_id=None,
               availability_zone_id=None, availability_zone=None):
        """Create new share group."""

        share_group_snapshot = None
        original_share_group = None
        # NOTE(gouthamr): share_server_id is inherited from the
        # parent share group if a share group snapshot is specified,
        # else, it will be set in the share manager.
        share_server_id = None
        if source_share_group_snapshot_id:
            share_group_snapshot = self.db.share_group_snapshot_get(
                context, source_share_group_snapshot_id)
            if share_group_snapshot['status'] != constants.STATUS_AVAILABLE:
                msg = (_("Share group snapshot status must be %s.")
                       % constants.STATUS_AVAILABLE)
                raise exception.InvalidShareGroupSnapshot(reason=msg)

            original_share_group = self.db.share_group_get(
                context, share_group_snapshot['share_group_id'])
            share_type_ids = [
                s['share_type_id']
                for s in original_share_group['share_types']]
            share_network_id = original_share_group['share_network_id']
            share_server_id = original_share_group['share_server_id']
            availability_zone_id = original_share_group['availability_zone_id']

        # Get share_type_objects
        share_type_objects = []
        driver_handles_share_servers = None
        for share_type_id in (share_type_ids or []):
            try:
                share_type_object = share_types.get_share_type(
                    context, share_type_id)
            except exception.ShareTypeNotFound:
                msg = _("Share type with id %s could not be found.")
                raise exception.InvalidInput(msg % share_type_id)
            share_type_objects.append(share_type_object)

            extra_specs = share_type_object.get('extra_specs')
            if extra_specs:
                share_type_handle_ss = strutils.bool_from_string(
                    extra_specs.get(
                        constants.ExtraSpecs.DRIVER_HANDLES_SHARE_SERVERS))
                if driver_handles_share_servers is None:
                    driver_handles_share_servers = share_type_handle_ss
                elif not driver_handles_share_servers == share_type_handle_ss:
                    # NOTE(ameade): if the share types have conflicting values
                    #  for driver_handles_share_servers then raise bad request
                    msg = _("The specified share_types cannot have "
                            "conflicting values for the "
                            "driver_handles_share_servers extra spec.")
                    raise exception.InvalidInput(reason=msg)

                if (not share_type_handle_ss) and share_network_id:
                    msg = _("When using a share types with the "
                            "driver_handles_share_servers extra spec as "
                            "False, a share_network_id must not be provided.")
                    raise exception.InvalidInput(reason=msg)

        share_network = {}
        try:
            if share_network_id:
                share_network = self.db.share_network_get(
                    context, share_network_id)
        except exception.ShareNetworkNotFound:
            msg = _("The specified share network does not exist.")
            raise exception.InvalidInput(reason=msg)

        if share_network:
            # Check if share network is active, otherwise raise a BadRequest
            api_common.check_share_network_is_active(share_network)

        if (driver_handles_share_servers and
                not (source_share_group_snapshot_id or share_network_id)):
            msg = _("When using a share type with the "
                    "driver_handles_share_servers extra spec as "
                    "True, a share_network_id must be provided.")
            raise exception.InvalidInput(reason=msg)

        try:
            share_group_type = self.db.share_group_type_get(
                context, share_group_type_id)
        except exception.ShareGroupTypeNotFound:
            msg = _("The specified share group type %s does not exist.")
            raise exception.InvalidInput(reason=msg % share_group_type_id)

        supported_share_types = set(
            [x['share_type_id'] for x in share_group_type['share_types']])
        supported_share_type_objects = [
            share_types.get_share_type(context, share_type_id) for
            share_type_id in supported_share_types
        ]

        if not set(share_type_ids or []) <= supported_share_types:
            msg = _("The specified share types must be a subset of the share "
                    "types supported by the share group type.")
            raise exception.InvalidInput(reason=msg)

        # Grab share type AZs for scheduling
        share_types_of_new_group = (
            share_type_objects or supported_share_type_objects
        )
        stype_azs_of_new_group = []
        stypes_unsupported_in_az = []
        for stype in share_types_of_new_group:
            stype_azs = stype.get('extra_specs', {}).get(
                'availability_zones', '')
            if stype_azs:
                stype_azs = stype_azs.split(',')
                stype_azs_of_new_group.extend(stype_azs)
                if availability_zone and availability_zone not in stype_azs:
                    # If an AZ is requested, it must be supported by the AZs
                    # configured in each of the share types requested
                    stypes_unsupported_in_az.append((stype['name'],
                                                     stype['id']))

        if stypes_unsupported_in_az:
            msg = _("Share group cannot be created since the following share "
                    "types are not supported within the availability zone "
                    "'%(az)s': (%(stypes)s)")
            payload = {'az': availability_zone, 'stypes': ''}
            for type_name, type_id in set(stypes_unsupported_in_az):
                if payload['stypes']:
                    payload['stypes'] += ', '
                type_name = '%s ' % (type_name or '')
            payload['stypes'] += type_name + '(ID: %s)' % type_id
            raise exception.InvalidInput(reason=msg % payload)

        try:
            reservations = QUOTAS.reserve(context, share_groups=1)
        except exception.OverQuota as e:
            overs = e.kwargs['overs']
            usages = e.kwargs['usages']
            quotas = e.kwargs['quotas']

            def _consumed(name):
                return (usages[name]['reserved'] + usages[name]['in_use'])

            if 'share_groups' in overs:
                msg = ("Quota exceeded for '%(s_uid)s' user in '%(s_pid)s' "
                       "project. (%(d_consumed)d of "
                       "%(d_quota)d already consumed).")
                LOG.warning(msg, {
                    's_pid': context.project_id,
                    's_uid': context.user_id,
                    'd_consumed': _consumed('share_groups'),
                    'd_quota': quotas['share_groups'],
                })
            raise exception.ShareGroupsLimitExceeded()

        options = {
            'share_group_type_id': share_group_type_id,
            'source_share_group_snapshot_id': source_share_group_snapshot_id,
            'share_network_id': share_network_id,
            'share_server_id': share_server_id,
            'availability_zone_id': availability_zone_id,
            'name': name,
            'description': description,
            'user_id': context.user_id,
            'project_id': context.project_id,
            'status': constants.STATUS_CREATING,
            'share_types': share_type_ids or supported_share_types
        }
        if original_share_group:
            options['host'] = original_share_group['host']

        share_group = {}
        try:
            share_group = self.db.share_group_create(context, options)
            if share_group_snapshot:
                members = self.db.share_group_snapshot_members_get_all(
                    context, source_share_group_snapshot_id)
                for member in members:
                    share_instance = self.db.share_instance_get(
                        context, member['share_instance_id'])
                    share_type = share_types.get_share_type(
                        context, share_instance['share_type_id'])
                    self.share_api.create(
                        context,
                        member['share_proto'],
                        member['size'],
                        None,
                        None,
                        share_group_id=share_group['id'],
                        share_group_snapshot_member=member,
                        share_type=share_type,
                        availability_zone=availability_zone_id,
                        share_network_id=share_network_id)
        except Exception:
            with excutils.save_and_reraise_exception():
                if share_group:
                    self.db.share_group_destroy(
                        context.elevated(), share_group['id'])
                QUOTAS.rollback(context, reservations)

        try:
            QUOTAS.commit(context, reservations)
        except Exception:
            with excutils.save_and_reraise_exception():
                QUOTAS.rollback(context, reservations)

        request_spec = {'share_group_id': share_group['id']}
        request_spec.update(options)
        request_spec['availability_zones'] = set(stype_azs_of_new_group)
        request_spec['share_types'] = share_type_objects
        request_spec['resource_type'] = share_group_type

        if share_group_snapshot and original_share_group:
            self.share_rpcapi.create_share_group(
                context, share_group, original_share_group['host'])
        else:
            self.scheduler_rpcapi.create_share_group(
                context, share_group_id=share_group['id'],
                request_spec=request_spec, filter_properties={})

        return share_group

    def delete(self, context, share_group):
        """Delete share group."""

        share_group_id = share_group['id']
        if not share_group['host']:
            self.db.share_group_destroy(context.elevated(), share_group_id)
            return

        statuses = (constants.STATUS_AVAILABLE, constants.STATUS_ERROR)
        if not share_group['status'] in statuses:
            msg = (_("Share group status must be one of %(statuses)s")
                   % {"statuses": statuses})
            raise exception.InvalidShareGroup(reason=msg)

        # NOTE(ameade): check for group_snapshots in the group
        if self.db.count_share_group_snapshots_in_share_group(
                context, share_group_id):
            msg = (_("Cannot delete a share group with snapshots"))
            raise exception.InvalidShareGroup(reason=msg)

        # NOTE(ameade): check for shares in the share group
        if self.db.count_shares_in_share_group(context, share_group_id):
            msg = (_("Cannot delete a share group with shares"))
            raise exception.InvalidShareGroup(reason=msg)

        share_group = self.db.share_group_update(
            context, share_group_id, {'status': constants.STATUS_DELETING})

        try:
            reservations = QUOTAS.reserve(
                context,
                share_groups=-1,
                project_id=share_group['project_id'],
                user_id=share_group['user_id'],
            )
        except exception.OverQuota as e:
            reservations = None
            LOG.exception(
                ("Failed to update quota for deleting share group: %s"), e)

        try:
            self.share_rpcapi.delete_share_group(context, share_group)
        except Exception:
            with excutils.save_and_reraise_exception():
                QUOTAS.rollback(context, reservations)

        if reservations:
            QUOTAS.commit(
                context, reservations,
                project_id=share_group['project_id'],
                user_id=share_group['user_id'],
            )

    def update(self, context, group, fields):
        return self.db.share_group_update(context, group['id'], fields)

    def get(self, context, share_group_id):
        return self.db.share_group_get(context, share_group_id)

    def get_all(self, context, detailed=True, search_opts=None, sort_key=None,
                sort_dir=None):

        if search_opts is None:
            search_opts = {}

        LOG.debug("Searching for share_groups by: %s",
                  six.text_type(search_opts))

        # Get filtered list of share_groups
        if search_opts.pop('all_tenants', 0) and context.is_admin:
            share_groups = self.db.share_group_get_all(
                context, detailed=detailed, filters=search_opts,
                sort_key=sort_key, sort_dir=sort_dir)
        else:
            share_groups = self.db.share_group_get_all_by_project(
                context, context.project_id, detailed=detailed,
                filters=search_opts, sort_key=sort_key, sort_dir=sort_dir)

        return share_groups

    def create_share_group_snapshot(self, context, name=None, description=None,
                                    share_group_id=None):
        """Create new share group snapshot."""
        options = {
            'share_group_id': share_group_id,
            'name': name,
            'description': description,
            'user_id': context.user_id,
            'project_id': context.project_id,
            'status': constants.STATUS_CREATING,
        }
        share_group = self.db.share_group_get(context, share_group_id)
        # Check status of group, must be active
        if not share_group['status'] == constants.STATUS_AVAILABLE:
            msg = (_("Share group status must be %s")
                   % constants.STATUS_AVAILABLE)
            raise exception.InvalidShareGroup(reason=msg)

        # Create members for every share in the group
        shares = self.db.share_get_all_by_share_group_id(
            context, share_group_id)

        # Check status of all shares, they must be active in order to snap
        # the group
        for s in shares:
            if not s['status'] == constants.STATUS_AVAILABLE:
                msg = (_("Share %(s)s in share group must have status "
                         "of %(status)s in order to create a group snapshot")
                       % {"s": s['id'],
                          "status": constants.STATUS_AVAILABLE})
                raise exception.InvalidShareGroup(reason=msg)

        try:
            reservations = QUOTAS.reserve(context, share_group_snapshots=1)
        except exception.OverQuota as e:
            overs = e.kwargs['overs']
            usages = e.kwargs['usages']
            quotas = e.kwargs['quotas']

            def _consumed(name):
                return (usages[name]['reserved'] + usages[name]['in_use'])

            if 'share_group_snapshots' in overs:
                msg = ("Quota exceeded for '%(s_uid)s' user in '%(s_pid)s' "
                       "project. (%(d_consumed)d of "
                       "%(d_quota)d already consumed).")
                LOG.warning(msg, {
                    's_pid': context.project_id,
                    's_uid': context.user_id,
                    'd_consumed': _consumed('share_group_snapshots'),
                    'd_quota': quotas['share_group_snapshots'],
                })
            raise exception.ShareGroupSnapshotsLimitExceeded()

        snap = {}
        try:
            snap = self.db.share_group_snapshot_create(context, options)
            members = []
            for s in shares:
                member_options = {
                    'share_group_snapshot_id': snap['id'],
                    'user_id': context.user_id,
                    'project_id': context.project_id,
                    'status': constants.STATUS_CREATING,
                    'size': s['size'],
                    'share_proto': s['share_proto'],
                    'share_instance_id': s.instance['id']
                }
                member = self.db.share_group_snapshot_member_create(
                    context, member_options)
                members.append(member)

            # Cast to share manager
            self.share_rpcapi.create_share_group_snapshot(
                context, snap, share_group['host'])
        except Exception:
            with excutils.save_and_reraise_exception():
                # This will delete the snapshot and all of it's members
                if snap:
                    self.db.share_group_snapshot_destroy(context, snap['id'])
                QUOTAS.rollback(context, reservations)

        try:
            QUOTAS.commit(context, reservations)
        except Exception:
            with excutils.save_and_reraise_exception():
                QUOTAS.rollback(context, reservations)

        return snap

    def delete_share_group_snapshot(self, context, snap):
        """Delete share group snapshot."""
        snap_id = snap['id']
        statuses = (constants.STATUS_AVAILABLE, constants.STATUS_ERROR)
        share_group = self.db.share_group_get(context, snap['share_group_id'])
        if not snap['status'] in statuses:
            msg = (_("Share group snapshot status must be one of"
                     " %(statuses)s") % {"statuses": statuses})
            raise exception.InvalidShareGroupSnapshot(reason=msg)

        self.db.share_group_snapshot_update(
            context, snap_id, {'status': constants.STATUS_DELETING})

        try:
            reservations = QUOTAS.reserve(
                context,
                share_group_snapshots=-1,
                project_id=snap['project_id'],
                user_id=snap['user_id'],
            )
        except exception.OverQuota as e:
            reservations = None
            LOG.exception(
                ("Failed to update quota for deleting share group snapshot: "
                 "%s"), e)

        # Cast to share manager
        self.share_rpcapi.delete_share_group_snapshot(
            context, snap, share_group['host'])

        if reservations:
            QUOTAS.commit(
                context, reservations,
                project_id=snap['project_id'],
                user_id=snap['user_id'],
            )

    def update_share_group_snapshot(self, context, share_group_snapshot,
                                    fields):
        return self.db.share_group_snapshot_update(
            context, share_group_snapshot['id'], fields)

    def get_share_group_snapshot(self, context, snapshot_id):
        return self.db.share_group_snapshot_get(context, snapshot_id)

    def get_all_share_group_snapshots(self, context, detailed=True,
                                      search_opts=None, sort_key=None,
                                      sort_dir=None):
        if search_opts is None:
            search_opts = {}
        LOG.debug("Searching for share group snapshots by: %s",
                  six.text_type(search_opts))

        # Get filtered list of share group snapshots
        if search_opts.pop('all_tenants', 0) and context.is_admin:
            share_group_snapshots = self.db.share_group_snapshot_get_all(
                context, detailed=detailed, filters=search_opts,
                sort_key=sort_key, sort_dir=sort_dir)
        else:
            share_group_snapshots = (
                self.db.share_group_snapshot_get_all_by_project(
                    context, context.project_id, detailed=detailed,
                    filters=search_opts, sort_key=sort_key, sort_dir=sort_dir,
                )
            )
        return share_group_snapshots

    def get_all_share_group_snapshot_members(self, context,
                                             share_group_snapshot_id):
        members = self.db.share_group_snapshot_members_get_all(
            context, share_group_snapshot_id)
        return members



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Add_share_type_projects

Revision ID: ef0c02b4366
Revises: 17115072e1c3
Create Date: 2015-02-20 10:49:40.744974

"""

# revision identifiers, used by Alembic.
revision = 'ef0c02b4366'
down_revision = '59eb64046740'

from alembic import op
from oslo_log import log
import sqlalchemy as sql


LOG = log.getLogger(__name__)


def upgrade():
    meta = sql.MetaData()
    meta.bind = op.get_bind()
    is_public = sql.Column('is_public', sql.Boolean)

    try:
        op.add_column('share_types', is_public)

        share_types = sql.Table('share_types', meta, is_public.copy())
        # pylint: disable=no-value-for-parameter
        share_types.update().values(is_public=True).execute()
    except Exception:
        LOG.error("Column |%s| not created!", repr(is_public))
        raise

    try:
        op.create_table(
            'share_type_projects',
            sql.Column('id', sql.Integer, primary_key=True, nullable=False),
            sql.Column('created_at', sql.DateTime),
            sql.Column('updated_at', sql.DateTime),
            sql.Column('deleted_at', sql.DateTime),
            sql.Column('share_type_id', sql.String(36),
                       sql.ForeignKey('share_types.id', name="stp_id_fk")),
            sql.Column('project_id', sql.String(length=255)),
            sql.Column('deleted', sql.Integer),
            sql.UniqueConstraint('share_type_id', 'project_id', 'deleted',
                                 name="stp_project_id_uc"),
            mysql_engine='InnoDB',
        )
    except Exception:
        LOG.error("Table |%s| not created!", 'share_type_projects')
        raise


def downgrade():
    try:
        op.drop_column('share_types', 'is_public')
    except Exception:
        LOG.error("share_types.is_public column not dropped")
        raise

    try:
        op.drop_table('share_type_projects')
    except Exception:
        LOG.error("share_type_projects table not dropped")
        raise



# Copyright (c) 2015 Mirantis Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt
from webob import exc

from manila.api.v2 import share_instance_export_locations as export_locations
from manila import context
from manila import db
from manila import exception
from manila import policy
from manila import test
from manila.tests.api import fakes
from manila.tests import db_utils


@ddt.ddt
class ShareInstanceExportLocationsAPITest(test.TestCase):

    def _get_request(self, version="2.9", use_admin_context=True):
        req = fakes.HTTPRequest.blank(
            '/v2/share_instances/%s/export_locations' % self.share_instance_id,
            version=version, use_admin_context=use_admin_context)
        return req

    def setUp(self):
        super(ShareInstanceExportLocationsAPITest, self).setUp()
        self.controller = (
            export_locations.ShareInstanceExportLocationController())
        self.resource_name = self.controller.resource_name
        self.ctxt = {
            'admin': context.RequestContext('admin', 'fake', True),
            'user': context.RequestContext('fake', 'fake'),
        }
        self.mock_policy_check = self.mock_object(
            policy, 'check_policy', mock.Mock(return_value=True))
        self.share = db_utils.create_share()
        self.share_instance_id = self.share.instance.id
        self.req = self._get_request()
        paths = ['fake1/1/', 'fake2/2', 'fake3/3']
        db.share_export_locations_update(
            self.ctxt['admin'], self.share_instance_id, paths, False)

    @ddt.data({'role': 'admin', 'version': '2.9'},
              {'role': 'user', 'version': '2.9'},
              {'role': 'admin', 'version': '2.13'},
              {'role': 'user', 'version': '2.13'})
    @ddt.unpack
    def test_list_and_show(self, role, version):

        summary_keys = ['id', 'path']
        admin_summary_keys = summary_keys + [
            'share_instance_id', 'is_admin_only']
        detail_keys = summary_keys + ['created_at', 'updated_at']
        admin_detail_keys = admin_summary_keys + ['created_at', 'updated_at']

        self._test_list_and_show(role, version, summary_keys, detail_keys,
                                 admin_summary_keys, admin_detail_keys)

    @ddt.data('admin', 'user')
    def test_list_and_show_with_preferred_flag(self, role):

        summary_keys = ['id', 'path', 'preferred']
        admin_summary_keys = summary_keys + [
            'share_instance_id', 'is_admin_only']
        detail_keys = summary_keys + ['created_at', 'updated_at']
        admin_detail_keys = admin_summary_keys + ['created_at', 'updated_at']

        self._test_list_and_show(role, '2.14', summary_keys, detail_keys,
                                 admin_summary_keys, admin_detail_keys)

    def _test_list_and_show(self, role, version, summary_keys, detail_keys,
                            admin_summary_keys, admin_detail_keys):

        req = self._get_request(version=version,
                                use_admin_context=(role == 'admin'))
        index_result = self.controller.index(req, self.share_instance_id)

        self.assertIn('export_locations', index_result)
        self.assertEqual(1, len(index_result))
        self.assertEqual(3, len(index_result['export_locations']))

        for index_el in index_result['export_locations']:
            self.assertIn('id', index_el)
            show_result = self.controller.show(
                req, self.share_instance_id, index_el['id'])
            self.assertIn('export_location', show_result)
            self.assertEqual(1, len(show_result))

            show_el = show_result['export_location']

            # Check summary keys in index result & detail keys in show result
            if role == 'admin':
                self.assertEqual(len(admin_summary_keys), len(index_el))
                for key in admin_summary_keys:
                    self.assertIn(key, index_el)
                self.assertEqual(len(admin_detail_keys), len(show_el))
                for key in admin_detail_keys:
                    self.assertIn(key, show_el)
            else:
                self.assertEqual(len(summary_keys), len(index_el))
                for key in summary_keys:
                    self.assertIn(key, index_el)
                self.assertEqual(len(detail_keys), len(show_el))
                for key in detail_keys:
                    self.assertIn(key, show_el)

            # Ensure keys common to index & show results have matching values
            for key in summary_keys:
                self.assertEqual(index_el[key], show_el[key])

    def test_list_export_locations_share_instance_not_found(self):
        self.assertRaises(
            exc.HTTPNotFound,
            self.controller.index,
            self.req, 'inexistent_share_instance_id',
        )

    def test_show_export_location_share_instance_not_found(self):
        index_result = self.controller.index(self.req, self.share_instance_id)
        el_id = index_result['export_locations'][0]['id']

        self.assertRaises(
            exc.HTTPNotFound,
            self.controller.show,
            self.req, 'inexistent_share_id', el_id,
        )

    @ddt.data('1.0', '2.0', '2.8')
    def test_list_with_unsupported_version(self, version):
        self.assertRaises(
            exception.VersionNotFoundForAPIMethod,
            self.controller.index,
            self._get_request(version),
            self.share_instance_id,
        )

    @ddt.data('1.0', '2.0', '2.8')
    def test_show_with_unsupported_version(self, version):
        index_result = self.controller.index(self.req, self.share_instance_id)

        self.assertRaises(
            exception.VersionNotFoundForAPIMethod,
            self.controller.show,
            self._get_request(version),
            self.share_instance_id,
            index_result['export_locations'][0]['id']
        )
